// Boid.h : Class for Boid object
// Animation Class CS527
// Hyejung Hur

#ifndef _BOID_H_
#define _BOID_H_

#include <pthread.h>
#include <vector>
#include "SimObject.h"

class Boid;

class BoidThreads {
public :
	int mId;
	pthread_t mthId;
	Boid* mBoid;
	BoidThreads() { };

};

class Boid : public SimObject {
public:
	enum boidTypes {
		NORMAL = 0,
		PREDATOR = 1,
		PREY = 2
	};

	enum methodTypes {
		BOIDS_MODE  = 0,
		TARGET_MODE = 1,
		AVOID_MODE  = 2,
		COORDINATE_MODE  = 3
	};

public:
	Boid(Ogre::SceneManager* mSceneMgr=NULL, char* name=NULL, char* filename=NULL, boidTypes eType=NORMAL); 
	Boid(Ogre::Vector3 bPosition, Ogre::Vector3 bVelocity, Ogre::Vector3 bDimensions, Ogre::SceneManager* mSceneMgr=NULL, char* name=NULL, char* filename=NULL,  boidTypes eType=NORMAL); 
	virtual ~Boid(void);

public:
	virtual bool update(const double &elapsedSeconds, bool renderOnly=false);  

public:
	void initThread(void);
	void destroyThread(void);
	static void* Boid::LoopThread(void *args);

public:
	void reset(void);
	static void setMethod(methodTypes type);

	void setTarget(Boid* target);
	void setSplitter(Boid* target);
	void setInitialVectors(Ogre::Vector3 pos, Ogre::Vector3 dir);

	void setMaxAcceleration(double acc);
	double setCruiseDistance(double distance);
	
	void setMaxVelocity(double velocity);
	double getMaxVelocity(void);

	void setFlockSelectively(bool flocking);
	Ogre::Vector3 getFlockCenter(void);


protected:
	void init(void);

	double accumulate(Ogre::Vector3 &accumulator, Ogre::Vector3 valueToAdd);
	Ogre::Vector3 navigator(bool renderOnly);
	Ogre::Vector3 wander(void);

	void calculateNeighbors(void);
	void calculateRollPitchYaw(Ogre::Vector3 appliedAcceleration, Ogre::Vector3 currentVelocity, Ogre::Vector3 currentPosition);
	int getClosestNeightbor(void);

	int visibleToSelf(Boid *b);
	Ogre::Vector3 collisionAvoidance(void);
	Ogre::Vector3 predatorAvoidance(void);
	Ogre::Vector3 maintainingCruisingDistance(void);
	Ogre::Vector3 velocityMatching(void);
	Ogre::Vector3 flockCentering(void);

	int getType(void);
	double getGravAcceleration(void);		
	float getProbeLength(void);
	double setSpeed(void);							
	
	int checkVisMatrix(Boid* n);
	void setVisMatrix(Boid* n, int flag);
	bool isRun(void);

protected:
	static std::vector<Boid*> m_boidList;
	static int m_boidCount;
	static int **m_visibilityMatrix;
	static methodTypes m_methodType;

 	std::vector<BoidThreads*> m_thIdList;

protected:
	boidTypes m_boidType;

	Boid* m_target;
	Boid* m_splitter;

	int m_boids_observed;
	Boid* m_closestNeighbor;  
	int m_foundClosestNeighbor;
	double m_distanceToClosestNeighbor;

	double m_bodyLength;
	double m_time;
	double m_timeout;
	bool m_start;

	bool m_visibility;
	bool m_flockSelectively;

	double m_steer;
	double m_maxVelocity;
	double m_maxAcceleration;
	double m_cruiseDistance;

	Ogre::Vector3 m_acceleration;
	Ogre::Vector3 m_flockcenter;
	Ogre::Vector3 m_initialPosition;
	Ogre::Vector3 m_initialVelocity;

};

#endif // _BOID_H_
