// Instancing.h : 
// Animation Class CS527
// Hyejung Hur

#include <OgreNoMemoryMacros.h>

#include <al.h>
#include <AL/alut.h>
#include <alc.h>

#include "CEGUI/CEGUI.h"
#include "OgreCEGUIRenderer.h"
#include "OgreConfigFile.h"
#include "OgreStringConverter.h"
#include "OgreException.h"
#include "OgreFrameListener.h"
#include "ExampleApplication.h"
#include "OgreInstancedGeometry.h"

#include "Boid.h"
#include "Obstacle.h"

#ifndef MAXBOIDS
#define MAXBOIDS 100
#endif

using namespace Ogre;

class InstancingApplication;

CEGUI::MouseButton convertOISMouseButtonToCegui(int buttonID)
{
  switch (buttonID)
    {
	case 0: return CEGUI::LeftButton;
	case 1: return CEGUI::RightButton;
	case 2:	return CEGUI::MiddleButton;
	case 3: return CEGUI::X1Button;
	default: return CEGUI::LeftButton;
    }
}
//out of the listener class...if set as a member, the sample crashes when moving the mouse!

// Event handler to add ability to alter subdivision
class InstancingListener : public ExampleFrameListener, public OIS::MouseListener, public OIS::KeyListener
{
protected:
	InstancingApplication * mMain;
	SceneManager* mSceneMgr;
	// the ray scene query pointer
	RaySceneQuery* mRaySceneQuery;     

	bool mRequestShutDown;
	bool mLMBDown;
	bool mRMBDown;
	int  mGrabbedEntity;
	Obstacle* mGrabbedEntityObj;

	double mAvgFrameTime;
	String mDebugText;

	double mBoidsRate;
	double  mlastUpdate;

	std::vector<Boid*> mBoidsList;
	std::vector<Boid*> mBoidsReadyList;

	std::vector<Obstacle*> mRangeBoxList;
	SceneNode* mBugNodeptr;
	Boid* m_preyptr;
	Boid* m_predatorptr;
	int m_showDimension;

	int mNUMBOIDS;
	int mDrawMode;
	int mSkipFrames;

	bool mKeyPressed;
	bool mStartStatus;

	int mRefreshCount;
	double mElapsedTime;
	CEGUI::Renderer* mGUIRenderer;

	CEGUI::Point mLastMousePosition;

	ALCcontext*	mpSoundContext;
	ALCdevice*	mpSoundDevice;
	unsigned int mALaSource;
	unsigned int mALSampleSet;
	bool mSoundPlay;

public:
	//----------------------------------------------------------------//
	InstancingListener(RenderWindow* win, Camera* cam,CEGUI::Renderer* renderer, InstancingApplication*main);
	~InstancingListener();

public:
	//----------------------------------------------------------------//
	void initalize(void);
	void start(void);
	void pause(void);
	void setBoidsCount(int count);
	void setBoidsCount(int count, Ogre::Vector3 pos, Vector3 attitude);
	void setMaxVelocity( double acc );
	double setCruiseDistance( double distance );
	void setMethod(int type);

	//----------------------------------------------------------------//
	bool frameStarted(const FrameEvent& evt);
	void requestShutdown(void);

	//----------------------------------------------------------------//
	bool handleMouseMove(const CEGUI::EventArgs& e);
	bool handleMouseButtonDown(const CEGUI::EventArgs& e);
	bool handleMouseButtonUp(const CEGUI::EventArgs& e);
	bool mouseMoved( const OIS::MouseEvent &arg );
	bool mousePressed( const OIS::MouseEvent &arg, OIS::MouseButtonID id );
	bool mouseReleased( const OIS::MouseEvent &arg, OIS::MouseButtonID id );
	void onLeftPressed(const OIS::MouseEvent &arg);

	//----------------------------------------------------------------//
	void updateStats(void);

	//----------------------------------------------------------------//
	virtual bool processUnbufferedKeyInput(const FrameEvent& evt);
	bool keyPressed( const OIS::KeyEvent &arg ) { return true; }
	bool keyReleased( const OIS::KeyEvent &arg ) { return true; }

protected:
	//----------------------------------------------------------------//
	int createBoids(void);
	void createBoundBox(void);
	void updateBoids(bool renderOnly);
	void getXY(double a, double b, double& x, double& y);
	Obstacle* getObstacle(const char* name);

	void loadSound(char* filename, bool looping);
};

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
class InstancingApplication : public ExampleApplication
{

public:
	//-----------------------------------------------------------------------
	InstancingApplication():mGUIRenderer(0),
        mGUISystem(0),
        mEditorGuiSheet(0),
		mStart(false) { }
	//-----------------------------------------------------------------------
	~InstancingApplication()
	{
		 if(mEditorGuiSheet)
       {
           CEGUI::WindowManager::getSingleton().destroyWindow(mEditorGuiSheet);
       }
       if(mGUISystem)
       {
           delete mGUISystem;
           mGUISystem = 0;
       }
       if(mGUIRenderer)
       {
           delete mGUIRenderer;
           mGUIRenderer = 0;
       }
	
	}
	RenderWindow*getRenderWindow(void)
	{
		return mWindow;
	}
   
	SceneManager * getSeceneManager(void) 
   {
	   return mSceneMgr;
   }

	void updatePropertyPanel(int idx, float value)
	{
		CEGUI::WindowManager& wmgr = CEGUI::WindowManager::getSingleton();
		CEGUI::Scrollbar* scrollptr= NULL;
		if(idx == 0)
		{
			scrollptr = ((CEGUI::Scrollbar*)(wmgr.getWindow((CEGUI::utf8*)"Object Count")));
			scrollptr->setScrollPosition(value);
		} else if (idx == 1)
		{
			scrollptr =((CEGUI::Scrollbar*)(wmgr.getWindow((CEGUI::utf8*)"Speed Burner")));
			scrollptr->setScrollPosition(value);
		} else if (idx == 2)
		{
			scrollptr =((CEGUI::Scrollbar*)(wmgr.getWindow((CEGUI::utf8*)"Avoid Distance")));
			scrollptr->setScrollPosition(value);
		}
	}

	void updatePropertyPanel(int type)
	{
		CEGUI::WindowManager& wmgr = CEGUI::WindowManager::getSingleton();
		CEGUI::RadioButton* buttonptr= NULL;
		if(type == 0)
		{
			// getSelectedButtonInGroup
			buttonptr = ((CEGUI::RadioButton*)(wmgr.getWindow((CEGUI::utf8*)"tBoidsMode")));
		} else if (type == 1)
		{
			buttonptr = ((CEGUI::RadioButton*)(wmgr.getWindow((CEGUI::utf8*)"tFlockMode")));
		} else if (type == 2)
		{
			buttonptr = ((CEGUI::RadioButton*)(wmgr.getWindow((CEGUI::utf8*)"tTargetMode")));
		} else if (type == 3)
		{
			buttonptr = ((CEGUI::RadioButton*)(wmgr.getWindow((CEGUI::utf8*)"tCollisionMode")));
		} else if (type == 4)
		{
			buttonptr = ((CEGUI::RadioButton*)(wmgr.getWindow((CEGUI::utf8*)"tCoordination")));
		}
		if (buttonptr)
			buttonptr->setSelected(true);

	}

protected:
   CEGUI::OgreCEGUIRenderer* mGUIRenderer;
   CEGUI::System* mGUISystem;
   CEGUI::Window* mEditorGuiSheet;
   InstancingListener* mInstanceLister;
   bool mStart;
   int mObjectCount;

	//----------------------------------------------------------------------
	// Just override the mandatory create scene method
	void createScene(void)
	{
		
		// Set ambient light
		mSceneMgr->setAmbientLight(ColourValue(0.2, 0.2, 0.2));
		Light* l = mSceneMgr->createLight("MainLight");
		//add a skybox
		mSceneMgr->setSkyBox(true, "Examples/StormySkyBox", 1000);
		//setup the light
		l->setType(Light::LT_DIRECTIONAL);
        Vector3 dir(-300,-1000,-500);
        dir.normalise();
		l->setDirection(dir);
		//l->setDirection(-0.5, -0.5, 0);
		
		/*l->setType(Light::LT_SPOTLIGHT);
        l->setPosition(1500,1550,500);
        l->setSpotlightRange(Degree(30), Degree(50));
        Vector3 dir;
        dir = -l->getPosition();
        dir.normalise();
        l->setDirection(dir);
        l->setDiffuseColour(0.8, 0.8, 0.8);
        l->setSpecularColour(0.9, 0.9, 1);*/

		//mCamera->setPosition(500,500, 1500);
		//mCamera->lookAt(0,0,0);
		mCamera->setPosition(100, 10, 400);
		mCamera->lookAt(100, 0, 0);
		Plane plane;
        plane.normal = Vector3::UNIT_Y;
        plane.d = 100;
        MeshManager::getSingleton().createPlane("Myplane",
            ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME, plane,
            1500,900,5,5,true,1,5,5,Vector3::UNIT_Z);
        Entity* pPlaneEnt = mSceneMgr->createEntity( "desktop", "Myplane" );
        pPlaneEnt->setMaterialName("Examples/Rockwall");
        pPlaneEnt->setCastShadows(false);
		//World* worldptr= mSceneMgr->getw->getWorld();

		pPlaneEnt->setQueryFlags(SimObject::WORLD_MASK);
        SceneNode* planeNode = mSceneMgr->getRootSceneNode()->createChildSceneNode();
		planeNode->attachObject(pPlaneEnt);
		planeNode->setPosition(0, 30, 0);

		SceneNode* winNode = mSceneMgr->getRootSceneNode()->createChildSceneNode();
		Entity* pWinEnt = mSceneMgr->createEntity( "window", "window.mesh" );
		winNode->setPosition(90, -70, -360);
		winNode->attachObject(pWinEnt);
	
		/*
		SceneNode* nodetest1  = mSceneMgr->getRootSceneNode()->createChildSceneNode();
		Entity* pnodeEnt1 = mSceneMgr->createEntity( "test1", "dog.mesh" );
		nodetest1->setPosition(-100, -70, -70);
		nodetest1->attachObject( pnodeEnt1 );

		SceneNode* nodetest2  = mSceneMgr->getRootSceneNode()->createChildSceneNode();
		Entity* pnodeEnt2 = mSceneMgr->createEntity( "test2", "penguine.mesh" );
		nodetest2->setPosition(100, -70, 50);
		nodetest2->attachObject( pnodeEnt2 );

		SceneNode* nodetest3  = mSceneMgr->getRootSceneNode()->createChildSceneNode();
		Entity* pnodeEnt3 = mSceneMgr->createEntity( "test3", "notes.mesh" );
		nodetest3->setPosition(250, -70, 150);
		nodetest3->attachObject( pnodeEnt3 );

		SceneNode* nodetest4  = mSceneMgr->getRootSceneNode()->createChildSceneNode();
		Entity* pnodeEnt4 = mSceneMgr->createEntity( "test4", "flower.mesh" );
		nodetest4->setPosition(-250, -70, -30);
		nodetest4->attachObject( pnodeEnt4 );


		*/


		setupGUI();
		setupEventHandlers();

		//set up the compositor
		Ogre::Viewport* pViewport = mWindow->getViewport(0);
		pViewport->setCamera(mCamera);
		CompositorManager::getSingleton().addCompositor(pViewport,"Bloom");

		
	}
	//-----------------------------------------------------------------------
	void destroyScene(void)
	{
	}
	//-----------------------------------------------------------------------
	void createFrameListener(void)
	{
		// This is where we instantiate our own frame listener
		mInstanceLister= new InstancingListener(mWindow, mCamera,mGUIRenderer,this);
		mFrameListener = mInstanceLister;
		mRoot->addFrameListener(mFrameListener);

		CEGUI::Window* wndw = CEGUI::WindowManager::getSingleton().getWindow("root");
		wndw->subscribeEvent(CEGUI::Window::EventMouseMove, CEGUI::Event::Subscriber(&InstancingListener::handleMouseMove,mInstanceLister));
		wndw->subscribeEvent(CEGUI::Window::EventMouseButtonUp, CEGUI::Event::Subscriber(&InstancingListener::handleMouseButtonUp,mInstanceLister));
		wndw->subscribeEvent(CEGUI::Window::EventMouseButtonDown, CEGUI::Event::Subscriber(&InstancingListener::handleMouseButtonDown,mInstanceLister));

		CEGUI::WindowManager& wmgr = CEGUI::WindowManager::getSingleton();
		CEGUI::Scrollbar* scrollptr=((CEGUI::Scrollbar*)(wmgr.getWindow((CEGUI::utf8*)"Object Count")));
		scrollptr->setScrollPosition(0.3);
		scrollptr =((CEGUI::Scrollbar*)(wmgr.getWindow((CEGUI::utf8*)"Speed Burner")));
		scrollptr->setScrollPosition(0.75);
		scrollptr =((CEGUI::Scrollbar*)(wmgr.getWindow((CEGUI::utf8*)"Avoid Distance")));
		scrollptr->setScrollPosition(0.015);
		mInstanceLister->initalize();
	}

	//-----------------------------------------------------------------------
	void setupResources(void)
	{
		ExampleApplication::setupResources();

		/*ResourceGroupManager *rsm = ResourceGroupManager::getSingletonPtr ();
		StringVector groups = rsm->getResourceGroups ();        
		if (std::find(groups.begin(), groups.end(), String("Instancing")) == groups.end())
		{
			rsm->createResourceGroup("Instancing");
			try
			{
				rsm->addResourceLocation("../../../../../ogreaddons/instancing/Media/materials/programs","FileSystem", "Instancing");
				rsm->addResourceLocation("../../../../../ogreaddons/instancing/Media/materials/scripts","FileSystem", "Instancing");
				rsm->addResourceLocation("../../../../../ogreaddons/instancing/Media/models","FileSystem", "Instancing");
			}
			catch (Ogre::Exception& e)
			{
				String error = e.getFullDescription();
				rsm->addResourceLocation("../../../Instancing/Media/materials/programs","FileSystem", "Instancing");
				rsm->addResourceLocation("../../../Instancing/Media/materials/scripts","FileSystem", "Instancing");
				rsm->addResourceLocation("../../../Instancing/Media/models","FileSystem", "Instancing");
			}
		}*/
	}
	void setupGUI(void)
	{
		// Set up GUI system
       mGUIRenderer = new CEGUI::OgreCEGUIRenderer(mWindow, Ogre::RENDER_QUEUE_OVERLAY, true, 3000, mSceneMgr);
       mGUISystem = new CEGUI::System(mGUIRenderer);
       CEGUI::Logger::getSingleton().setLoggingLevel(CEGUI::Informative);

	   CEGUI::SchemeManager::getSingleton().loadScheme((CEGUI::utf8*)"TaharezLookSkin.scheme");
       mGUISystem->setDefaultMouseCursor((CEGUI::utf8*)"TaharezLook", (CEGUI::utf8*)"MouseArrow");
       CEGUI::MouseCursor::getSingleton().setImage("TaharezLook", "MouseMoveCursor");
       mEditorGuiSheet= CEGUI::WindowManager::getSingleton().createWindow((CEGUI::utf8*)"DefaultWindow", (CEGUI::utf8*)"Sheet");  
       mGUISystem->setGUISheet(mEditorGuiSheet);
	 


	    mEditorGuiSheet = CEGUI::WindowManager::getSingleton().loadWindowLayout((CEGUI::utf8*)"InstancingDemo.layout");
		mGUISystem->setGUISheet(mEditorGuiSheet);
	}

	void setupEventHandlers(void)
	{
		using namespace CEGUI;
		CEGUI::WindowManager& wmgr = CEGUI::WindowManager::getSingleton();
		wmgr.getWindow((CEGUI::utf8*)"StartDemoBtn")->subscribeEvent(CEGUI::PushButton::EventClicked, CEGUI::Event::Subscriber(&InstancingApplication::handleStart, this));

		wmgr.getWindow((CEGUI::utf8*)"tBoidsMode")->subscribeEvent(CEGUI::RadioButton::EventSelectStateChanged, CEGUI::Event::Subscriber(&InstancingApplication::handleTechniqueChanged, this));
		wmgr.getWindow((CEGUI::utf8*)"tFlockMode")->subscribeEvent(CEGUI::RadioButton::EventSelectStateChanged, CEGUI::Event::Subscriber(&InstancingApplication::handleTechniqueChanged, this));
		wmgr.getWindow((CEGUI::utf8*)"tTargetMode")->subscribeEvent(CEGUI::RadioButton::EventSelectStateChanged, CEGUI::Event::Subscriber(&InstancingApplication::handleTechniqueChanged, this));
		wmgr.getWindow((CEGUI::utf8*)"tCollisionMode")->subscribeEvent(CEGUI::RadioButton::EventSelectStateChanged, CEGUI::Event::Subscriber(&InstancingApplication::handleTechniqueChanged, this));
		wmgr.getWindow((CEGUI::utf8*)"tCoordination")->subscribeEvent(CEGUI::RadioButton::EventSelectStateChanged, CEGUI::Event::Subscriber(&InstancingApplication::handleTechniqueChanged, this));

		wmgr.getWindow((CEGUI::utf8*)"Object Count")->subscribeEvent(CEGUI::Scrollbar::EventScrollPositionChanged,CEGUI::Event::Subscriber(&InstancingApplication::handleObjectCountChanged, this));
		wmgr.getWindow((CEGUI::utf8*)"Speed Burner")->subscribeEvent(CEGUI::Scrollbar::EventScrollPositionChanged,CEGUI::Event::Subscriber(&InstancingApplication::handleSpeedBurnerChanged, this));
		wmgr.getWindow((CEGUI::utf8*)"Avoid Distance")->subscribeEvent(CEGUI::Scrollbar::EventScrollPositionChanged,CEGUI::Event::Subscriber(&InstancingApplication::handleAvoidDistanceChanged, this));

		wmgr.getWindow((CEGUI::utf8*)"PostEffect")->subscribeEvent(CEGUI::Checkbox::EventCheckStateChanged,CEGUI::Event::Subscriber(&InstancingApplication::handlePostEffectChanged, this));
		wmgr.getWindow((CEGUI::utf8*)"Shadows")->subscribeEvent(CEGUI::Checkbox::EventCheckStateChanged,CEGUI::Event::Subscriber(&InstancingApplication::handleShadowsChanged, this));

	}
	//-----------------------------------------------------------------------
	bool handleStart(const CEGUI::EventArgs& e)
	{
		//static_cast<InstancingListener*>(mFrameListener)->requestShutdown();
		if(mStart == false) 
		{
			mStart = true;
			mInstanceLister->start();
			CEGUI::WindowManager::getSingleton().getWindow((CEGUI::utf8*)"StartDemoBtn")->setText(CEGUI::String("Pause"));
		} else 
		{
			mStart = false;
			mInstanceLister->pause();
			CEGUI::WindowManager::getSingleton().getWindow((CEGUI::utf8*)"StartDemoBtn")->setText(CEGUI::String("Start"));
		}
		return true;
	}

	//-----------------------------------------------------------------------
	bool handleTechniqueChanged(const CEGUI::EventArgs& e)
	{
		CEGUI::uint id = ((CEGUI::RadioButton*)((const CEGUI::WindowEventArgs&)e).window)->getSelectedButtonInGroup()->getID();
		mInstanceLister->setMethod(id);
		return true;
	}
	//-----------------------------------------------------------------------
	bool handleObjectCountChanged(const CEGUI::EventArgs& e)
	{
		//static_cast<InstancingListener*>(mFrameListener)->destroyCurrentGeomOpt();
		float scrollval = ((CEGUI::Scrollbar*)((const CEGUI::WindowEventArgs&)e).window)->getScrollPosition();
		int objectCount = 100*scrollval;
		mObjectCount = objectCount;			
		mInstanceLister->setBoidsCount(mObjectCount);
		//static_cast<InstancingListener*>(mFrameListener)->createCurrentGeomOpt();
		String value=StringConverter::toString(objectCount);
		CEGUI::WindowManager::getSingleton().getWindow((CEGUI::utf8*)"Object Count Number")->setText(CEGUI::String(value.c_str()));

		return true;
	}
	//-----------------------------------------------------------------------
	bool handleSpeedBurnerChanged(const CEGUI::EventArgs& e)
	{
		
		float scrollval = ((CEGUI::Scrollbar*)((const CEGUI::WindowEventArgs&)e).window)->getScrollPosition();
		double speedBurned = 100.0f *scrollval;
		mInstanceLister->setMaxVelocity(speedBurned);
		char* timeChar= new char[10];
		sprintf(timeChar,"%0.1f",speedBurned);
		CEGUI::String timeText(timeChar);
		CEGUI::WindowManager::getSingleton().getWindow("Speed Value")->setText(timeText);

		//static_cast<InstancingListener*>(mFrameListener)->setBurnedTime(timeBurned);
		delete[] timeChar;
		return true;
	}

	//-----------------------------------------------------------------------
	bool handleAvoidDistanceChanged(const CEGUI::EventArgs& e)
	{
		
		float scrollval = ((CEGUI::Scrollbar*)((const CEGUI::WindowEventArgs&)e).window)->getScrollPosition();
		double distance = 100.0*scrollval;
		double ret = mInstanceLister->setCruiseDistance(distance);
		if(distance != ret)
		{
			distance = ret * 0.01;
			((CEGUI::Scrollbar*)((const CEGUI::WindowEventArgs&)e).window)->setScrollPosition(distance);
		}

		char* timeChar= new char[10];
		sprintf(timeChar,"%0.1f",distance);
		CEGUI::String timeText(timeChar);
		CEGUI::WindowManager::getSingleton().getWindow("Avoid Value")->setText(timeText);

		//static_cast<InstancingListener*>(mFrameListener)->setBurnedTime(timeBurned);
		delete[] timeChar;
		return true;
	}

	//-----------------------------------------------------------------------
	bool handlePostEffectChanged(const CEGUI::EventArgs& e)
	{
		Ogre::Viewport* pViewport = mWindow->getViewport(0);
		if(((CEGUI::Checkbox*)((const CEGUI::WindowEventArgs&)e).window)->isSelected())
		{
				CompositorManager::getSingleton().setCompositorEnabled(pViewport,"Bloom",true);
		}
		else
		{
				CompositorManager::getSingleton().setCompositorEnabled(pViewport,"Bloom",false);
		}
		return true;
	}
	//-----------------------------------------------------------------------
	bool handleShadowsChanged(const CEGUI::EventArgs&e)
	{
		if(((CEGUI::Checkbox*)((const CEGUI::WindowEventArgs&)e).window)->isSelected())
		{	
			//SHADOWTYPE_TEXTURE_MODULATIVE
			// SHADOWTYPE_STENCIL_ADDITIVE
			mSceneMgr->setShadowTechnique(SHADOWTYPE_TEXTURE_MODULATIVE);
		}
		else
		{
			mSceneMgr->setShadowTechnique(SHADOWTYPE_NONE);
		}
		return true;
	}
	
};
