// Obstacle.h : class for obstacles 
// Animation Class CS527
// Hyejung Hur

#ifndef _OBSTACLE_H_
#define _OBSTACLE_H_

#include <iostream>
#include <vector>
#include "SimObject.h"

struct ISectData {
	int m_intersectionflag;			// 1 = collide, 0 = didn't
	Ogre::Vector3 m_normal;			// normal at intersection
	Ogre::Vector3 m_point;			// point of intersection
};

class Obstacle;
class ObstaclePolygon;
class ObstacleBox;
class ObstacleSphere;

/* ########################################################################
 OBSTACLE
######################################################################### */
class Obstacle : public SimObject {
public:
	Obstacle(void);
	Obstacle(Ogre::SceneManager* mSceneMgr, char* name, char* filename, char* anifilename, double gab);
	virtual ~Obstacle(void);
	
public:
	virtual Obstacle *clone(void) = 0;
	virtual bool update(const double &elapsedSeconds, bool renderOnly=false); 
	virtual void setPosition(Ogre::Vector3 newPosition); 

	int getEnergy(void);
	void setAniMaxAngle(double  angle);
	void addEnergy(void);
    
protected:
	virtual ISectData intersectionWithRay(Ogre::Vector3 & raydirection, Ogre::Vector3 &rayorigin)  = 0;
	
public:
	ISectData doesRayIntersect(Ogre::Vector3 raydirection, Ogre::Vector3 rayorigin);
	
public:
	static std::vector<Obstacle*> m_obstacleList;
	static int m_obstracleCount;

protected: 
	int m_energy;
	Ogre::Degree m_angle;
	double m_angleStep;
	double m_maxAngle;

	Ogre::SceneNode* m_enegyNode;
	Ogre::Entity* m_enegyEntity;

};

/* ########################################################################
 POLYGON
######################################################################### */
class ObstaclePolygon : public Obstacle {
public:
	ObstaclePolygon(int numverts, Ogre::Vector3** vList,Ogre::SceneManager* mSceneMgr=NULL, char* name=NULL, char* filename=NULL, int Id=-1);
	ObstaclePolygon(int numverts, Ogre::Vector3** vList, int Id=-1);
	ObstaclePolygon(const ObstaclePolygon &p, Ogre::SceneManager* mSceneMgr=NULL, char* name=NULL, char* filename=NULL); 
	virtual ~ObstaclePolygon(void);
 
public: 
	virtual Obstacle *clone(void);  

public: 
	int operator==(const ObstaclePolygon &a);  
	void update(void);	 
	Ogre::Vector3 ** getVertex(void);
	
protected:
	virtual ISectData intersectionWithRay(Ogre::Vector3 &raydirection, Ogre::Vector3 &rayorigin);
	void init(int numverts, Ogre::Vector3** vList, int Id);

private:
	ObstaclePolygon(void) { };

	float m_d;				
	int m_i1;
	int m_i2;				
	Ogre::Vector3 m_normal;		
	int m_numvertices;		
	Ogre::Vector3 **m_vertex;		

};

/* ########################################################################
 BOX
######################################################################### */
class ObstacleBox : public Obstacle {
public: 
	ObstacleBox(const Ogre::Vector3 &topLeftBackCorner, const Ogre::Vector3 &bottomRightFrontCorner, Ogre::SceneManager* mSceneMgr=NULL, char* name=NULL, char* filename=NULL);
	ObstacleBox(const Ogre::Vector3 &topLeftBackCorner, const Ogre::Vector3 &bottomRightFrontCorner);
	ObstacleBox(const ObstacleBox &b, Ogre::SceneManager* mSceneMgr=NULL, char* name=NULL, char* filename=NULL);
	ObstacleBox(Ogre::SceneManager* mSceneMgr, char* name, Ogre::Vector3 bPosition, Ogre::Vector3 bDimensions, char* filename= NULL, char* anifilename=NULL, double gab = 0.0f); 
  
	virtual ~ObstacleBox(void);

public: 
	virtual Obstacle *clone(void);  
	virtual void setPosition(Ogre::Vector3 newPosition); 
	virtual bool update(const double &elapsedSeconds, bool renderOnly=false); 

	int operator==(const ObstacleBox &a);  

protected:
	virtual ISectData intersectionWithRay(Ogre::Vector3 &raydirection, Ogre::Vector3 &rayorigin);

	void init(const Ogre::Vector3 &topLeftBackCorner, const Ogre::Vector3 &bottomRightFrontCorner);
	void update(int idx, Ogre::Vector3 ** vList, double width, double height, double length);

private:
	ObstacleBox(void) { };

	Ogre::Vector3 m_tlb; 
	Ogre::Vector3 m_brf;  

	ObstaclePolygon *m_side[6]; 
	ObstacleSphere *m_boundingSphere;

};

/* ########################################################################
 SPHERE
######################################################################### */
class ObstacleSphere : public Obstacle {
public:
	ObstacleSphere(const Ogre::Vector3 &sphereOrigin, const double sphereRadius, Ogre::SceneManager* mSceneMgr=NULL, char* name=NULL, char* filename=NULL, int Id=-1);    
	ObstacleSphere(const Ogre::Vector3 &sphereOrigin, const double sphereRadius, int Id=-1);    
	virtual ~ObstacleSphere(void);
	
public:
	virtual Obstacle *clone(void);
	int operator==(ObstacleSphere &a);  

	void update(const Ogre::Vector3 &sphereOrigin);

protected:
	virtual ISectData intersectionWithRay(Ogre::Vector3 &raydirection, Ogre::Vector3 &rayorigin);

	void init(const Ogre::Vector3 &sphereOrigin, const double sphereRadius, int Id);

private:
	ObstacleSphere(void) { };

	double m_radius;  
	Ogre::Vector3 m_origin;  

};

#endif	// _OBSTACLE_H_