// SimObject.h : Base class for all object components 
// Animation Class CS527
// Hyejung Hur

#ifndef _SIMOBJECT_H_
#define _SIMOBJECT_H_

#include <OgreSceneManager.h>
#include <OgreMeshManager.h>
#include <OgreEntity.h>
#include <string>

class SimObject {
public:
	enum {
		NORMAL_MODE = 0,
		DIMENSION_MODE = 1,
		ALL_SHOW_MODE = 2,
		ALL_HIDE_MODE = 3
	};
	enum QueryFlags
	{
		  MOVABLE_MASK  = 1<<0,
		  STATIC_MASK   = 1<<1,
		  OBSTACLE_MASK = 1<<2,
		  WORLD_MASK    = 1<<3
	};

public:
	SimObject(void);				     
	SimObject(Ogre::SceneManager* mSceneMgr, char* name, char* filename, char* anifilename= NULL, double gab = 0.0f);				     
	virtual ~SimObject(void);	

public:
	virtual bool update(double elapsedSeconds, bool renderOnly);  
	virtual void setPosition(Ogre::Vector3 newPosition); 

public:
	void initScene(bool update = true);

	void setId(int Id);
	int getId(void);
	const char* getName(void);

	Ogre::Vector3 &getPosition(void); 
	double getLastUpdate(void);
	
	void setDimensions(Ogre::Vector3 newDimensions); 
	Ogre::Vector3 &getDimensions(void);   
	
	void setMass(double newMass);
	double getMass(void);		   

	void setVelocity(Ogre::Vector3 newVelocity); 
	Ogre::Vector3 &getVelocity(void);  

	void setDimensionDrawMode(int mode);
	void resetTime(double time);

	bool isActive(void);
	void setActive(bool flag);
	void showCollide(bool flag);

protected:
	int m_id;
	double m_mass;

	Ogre::Vector3 m_velocity;
	Ogre::Vector3 m_oldVelocity;
	Ogre::Vector3 m_position;	
	Ogre::Vector3 m_dimensions;
	
	double m_roll;
	double m_pitch;
	double m_yaw;

	double m_lastUpdate;
  	std::string m_name;

	Ogre::SceneNode* m_node;
	Ogre::SceneNode* m_dimNode;
	Ogre::SceneNode* m_aniGeoNode;
	Ogre::SceneNode* m_geomNode;
	Ogre::Entity* m_dimEntity;
	Ogre::Entity* m_geomEntity;
	Ogre::Entity* m_aniGeomEntity;
	int m_drawMode;
	bool m_active;

private:
  
};

inline double SimObject::getLastUpdate(void)
{
	return m_lastUpdate;
}

inline void SimObject::resetTime(double time)
{
	m_lastUpdate = time;
}

inline const char* SimObject::getName(void)
{
	return m_name.c_str();
}

inline bool SimObject::isActive(void)
{
	return m_active;
}

inline void SimObject::setActive(bool flag)
{
	m_active= flag;
}

inline void SimObject::showCollide(bool flag)
{
	if(flag == true)
	{
		m_dimEntity->setMaterialName("red_dark");
	} else 
	{
		m_dimEntity->setMaterialName("blue_dark");
	}
}

inline bool SimObject::update(double elapsedSeconds, bool renderOnly) 
{
	m_lastUpdate = elapsedSeconds;
	if(m_node != NULL) {
		m_node->setPosition(m_position.x, m_position.y, m_position.z);
		//m_node->translate(m_position.x, m_position.y, m_position.z);
		m_node->yaw(Ogre::Degree(m_yaw));
	}

	return true;
}

inline void SimObject::setId(int Id) 
{
	m_id = Id;
}	

inline int SimObject::getId(void) 
{
	return m_id;
}

inline void SimObject::setPosition(Ogre::Vector3 newPosition) 
{
	m_position = newPosition;
}

inline Ogre::Vector3 &SimObject::getPosition(void) 
{
	return m_position;
}

inline void SimObject::setDimensions(Ogre::Vector3 newDimensions) 
{
  m_dimensions = newDimensions;
}

inline Ogre::Vector3 &SimObject::getDimensions(void) 
{
	return m_dimensions;
}

inline void SimObject::setMass(double newMass) 
{
	m_mass = newMass;
}

inline double SimObject::getMass(void) 
{
	return m_mass;
}

inline void SimObject::setVelocity(Ogre::Vector3 newVelocity) 
{
	m_velocity = newVelocity;
}

inline Ogre::Vector3 &SimObject::getVelocity(void) 
{
	return m_velocity;
}

Ogre::Vector3 direction(Ogre::Vector3 &a);
Ogre::Vector3 direction(const Ogre::Vector3 &a);
Ogre::Vector3 direction(const double &x, const double &y, const double &z);
inline Ogre::Vector3 direction(Ogre::Vector3 &a) {
	Ogre::Vector3 u = a;
	u.normalise();
	return(u);
}
inline Ogre::Vector3 direction(const Ogre::Vector3 &a) {
	Ogre::Vector3 u = a;
	u.normalise();
	return(u);
}
inline Ogre::Vector3 direction(const double &x, const double &y, const double &z) {
	Ogre::Vector3 u;
	u.set(x, y, z);
	return direction(u);
}

#endif // _SIMOBJECT_H_









