// Vector.h :  Class for 3-vectors
// Animation Class CS527
// Hyejung Hur

#ifndef _VECTOR_H_
#define _VECTOR_H_

#include <math.h>
#include <iostream>

#ifndef DBL_MAX
#define DBL_MAX 1.7e38	
#endif

class Vector;

class Vector {
public:
	double x, y, z;                       
  
public:
	Vector(void);                           
	Vector(double a, double b, double c);  

public:  
	void set(double a, double b, double c);  
	void setDirection(Vector &d);     
	void copyDirection(Vector &d);     
	void setMagnitude(double m);      
	void copyMagnitude(Vector &v);     
	void normalize(void);                   
	double length(void);					 
	void negate(void);
	
	int isSame(const Vector &a);
	Vector &operator=(Vector &b);                              
	Vector &operator=(const Vector &b);                              
	double operator[](int index);    
	double get(int index);    

	                       
	int operator!=(Vector &a);    
	int operator==(Vector &a);    
	Vector operator+(const Vector &a);   
	Vector operator-(const Vector &a);  
	Vector &operator+=(const Vector &a);   
	Vector &operator-=(const Vector &a);  
	                            
	Vector operator%(const Vector &a);	// cross product
	double dot(Vector &a);
	double operator*(const Vector &a);	// dot product
	
	Vector operator*(const double &a);   
	Vector &operator*=(const double &b);    
	        
	Vector operator/(const double &a);   
	Vector &operator/=(const double &a);            

	double magnitude(void); 
	double magnitude(const Vector &a);                       

	double angleBetween(const Vector &a, const Vector &b);
	double angleBetween(const Vector &a);

};

Vector direction(Vector &a);
Vector direction(const Vector &a);
Vector direction(const double &x, const double &y, const double &z);
inline Vector direction(Vector &a) {
	Vector u = a;
	u.normalize();
	return(u);
}
inline Vector direction(const Vector &a) {
	Vector u = a;
	u.normalize();
	return(u);
}
inline Vector direction(const double &x, const double &y, const double &z) {
	Vector u;
	u.set(x, y, z);
	return direction(u);
}

inline std::ostream &operator<<(std::ostream &strm, const Vector &v) {
	return strm << "[" << v.x << ", " << v.y << ", " << v.z << "]";
}

#ifndef VLENSQRD
#define VLENSQRD(a, b, c) ((a)*(a) + (b)*(b) + (c)*(c))
#endif

#ifndef VLEN
#define VLEN(a, b, c) sqrt(VLENSQRD(a, b, c))
#endif

inline Vector::Vector(void) :
x(0), y(0), z(0) { 
}         

inline Vector::Vector( double a, double b, double c) :
x(a), y(b), z(c) { 
} 

inline void Vector::set(double a, double b, double c) {
	x = a;
	y = b;
	z = c;
}

inline void Vector::normalize(void) {
	double mag = VLEN(x, y, z);

	if (mag == 0) return;

	x /= mag;
	y /= mag;
	z /= mag;
 }

inline void Vector::setMagnitude(double m) {
	this->normalize();
	*this *= m;
}

inline void Vector::copyMagnitude(Vector &v) {
	setMagnitude(magnitude(v));
}

inline void Vector::setDirection(Vector &d) {
  double m = magnitude(*this);
  Vector v = d;
  
  v.normalize();
  Vector temp = v * m;
  *this = temp; 
}

inline void Vector::copyDirection(Vector &d) {
	//setDirection(direction(d));
}

inline double Vector::length(void) {
	return magnitude(*this);
}

inline void Vector::negate(void)
{
	x *= -1;
	y *= -1;
	z *= -1;
}

inline Vector &Vector::operator=(Vector &b) { 
	x = b.x; 
	y = b.y; 
	z = b.z; 

	return(*this);
}

inline Vector &Vector::operator=(const Vector &b) { 
	x = b.x; 
	y = b.y; 
	z = b.z; 

	return(*this);
}
 
inline double Vector::get(int index) {  
	if (index == 0) 
		return(x);
	else if (index == 1) 
		return(y);
	else if (index == 2) 
		return(z);
	else {
		std::cerr << "WARNING: You're using subscripting to access a nonexistant vector component (one other than x, y, or z). THIS IS BAD!!\n";
		//exit(-1);
		return(z); // this will never get called, but prevents compiler warnings...
	}
}
	
inline double Vector::operator[](int index) {  
	if (index == 0) 
		return(x);
	else if (index == 1) 
		return(y);
	else if (index == 2) 
		return(z);
	else {
		std::cerr << "WARNING: You're using subscripting to access a nonexistant vector component (one other than x, y, or z). THIS IS BAD!!\n";
		//exit(-1);
		return(z); // this will never get called, but prevents compiler warnings...
	}
}

inline int Vector::operator!=(Vector &a) {
	return(x != a.x || y != a.y || z != a.z);
}

inline int Vector::isSame(const Vector &a)
{
	return(x == a.x && y == a.y && z == a.z);
}
	
inline int Vector::operator==(Vector &a) { 
	return(x == a.x && y == a.y && z == a.z);
}

inline Vector Vector::operator+(const Vector &a) {
	Vector c(x+a.x, y+a.y, z+a.z);
	return(c);
}
	
inline Vector Vector::operator-(const Vector &a) { 
	Vector c(x-a.x, y-a.y, z-a.z);
	return(c);		
}

inline Vector &Vector::operator+=(const Vector &a) {
	x += a.x;
	y += a.y;
	z += a.z;
	return(*this);
}

inline Vector &Vector::operator-=(const Vector &a) { 
	x -= a.x;
	y -= a.y;
	z -= a.z;
	return(*this);
}

inline Vector Vector::operator%(const Vector &a) { 
	Vector c(y*a.z - z*a.y, z*a.x - x*a.z, x*a.y - y*a.x);
	return(c);
}

inline double Vector::dot(Vector &a) { 
	return(x*a.x + y*a.y + z*a.z);
}

inline double Vector::operator*(const Vector &a) { 
	return(x*a.x + y*a.y + z*a.z);
}
	
inline Vector Vector::operator*(const double &a) { 
	Vector c(x*a, y*a, z*a);
	return(c);	
}

inline Vector &Vector::operator*=(const double &a) { 
	x *= a;
	y *= a;
	z *= a;
	return(*this);
}

inline Vector Vector::operator/(const double &a) {
	if (a == 0) 
		std::cerr << "WARNING: You're dividing a vector by a zero-length scalar! NOT GOOD!\n";

	Vector c(x*1/a, y*1/a, z*1/a);
	return(c);
}

inline Vector &Vector::operator/=(const double &a) {
	x = x/a;
	y = y/a;
	z = z/a;
	
 	return(*this);
}

inline double Vector::magnitude(void) {
	return VLEN(x, y, z);
}

inline double Vector::magnitude(const Vector &a) {
	return VLEN(a.x, a.y, a.z);
}

inline double Vector::angleBetween(const Vector &a, const Vector &b) {
  // returns angle between a and b in RADIANS
  Vector c(a.x, a.y, a.z);
  double angle = c * b;
  
  return acos( angle / (magnitude(a) * magnitude(b) ) );
} 

inline double Vector::angleBetween(const Vector &a) {
  // returns angle between a and b in RADIANS
  Vector c(x, y, z);
  double angle = c * a;
  
  return acos( angle / (magnitude(a) * magnitude(c) ) );
} 

#endif // _VECTOR_H_
