// Boid.cpp : Class for Boid object
// Animation Class CS527
// Hyejung Hur

#include <Windows.h>
#include "Boid.h"
#include "Obstacle.h"
#include "BoidTypes.h"

using namespace std;
using namespace Ogre;

vector<Boid*> Boid::m_boidList;
int Boid::m_boidCount =0;
int ** Boid::m_visibilityMatrix;
Boid::methodTypes Boid::m_methodType = BOIDS_MODE;


Boid::Boid(SceneManager* mSceneMgr, char* name, char* filename, boidTypes eType) :
m_boidType(eType), m_flockSelectively(false), m_visibility(true), m_steer(0.0f),
m_maxVelocity(10.0), m_maxAcceleration(0.5), SimObject(mSceneMgr, name, filename) 
{
	m_bodyLength = 0;
	m_cruiseDistance = 0;
	m_velocity.set(0,0,0);

	init();
	initScene();
}
	
Boid::Boid(Vector3 bPosition,Vector3 bVelocity, Vector3 bDimensions, SceneManager* mSceneMgr, char* name, char* filename, boidTypes eType) :
m_boidType(eType), m_flockSelectively(false), m_visibility(true), m_steer(0.0f),
m_maxVelocity(10.0), m_maxAcceleration(0.5), SimObject(mSceneMgr, name, filename)  
{
	m_position = bPosition;
	m_initialPosition= m_position;
	m_velocity = bVelocity;
	m_initialVelocity= m_position;
	// width, height, length
	m_dimensions = bDimensions; 
	m_bodyLength = bDimensions.z;

	m_cruiseDistance = 2 * m_dimensions.z; 

	init();
	initScene();
}		
  
Boid::~Boid(void) 
{
	if(m_visibilityMatrix != NULL)
	{
		m_boidList.clear();
		
		for (int i = 0; i < m_boidCount-1; i++) {
			delete[] m_visibilityMatrix[i];
		}
		delete[] m_visibilityMatrix;
		m_visibilityMatrix = NULL;
	}
}
 

void Boid::init(void)
{
	m_oldVelocity.set(0,0,0);
	m_acceleration.set(0,0,0);
	m_flockcenter.set(0,0,0);
	m_target = NULL;
	m_splitter = NULL;

	m_id = m_boidList.size();
	char bName[10];
	sprintf(bName, "boid%d", m_id);
	m_name = bName;

	m_boidList.push_back(this);
	m_boidCount++;	
	cout << m_name << " is created " << endl;

	int **newMatrix = new int*[m_boidCount];
	for (int i = 0; i < m_boidCount; i++) {
		newMatrix[i] = new int[m_boidCount];
	}

	for (int i = 0; i < m_boidCount-1; i++) {
		for (int j = 0; j < m_boidCount-1; j++) {
			newMatrix[i][j] = m_visibilityMatrix[i][j];
		}
	}

	int **temp = m_visibilityMatrix;
	m_visibilityMatrix = newMatrix;

	for (int i = 0; i < m_boidCount-1; i++) {
		delete[] temp[i];
	}
	delete[] temp;

	QueryFlags mask = STATIC_MASK;
	if(m_boidType != NORMAL)
		mask = MOVABLE_MASK;
	
	if(m_geomEntity != NULL)
	{
		m_geomEntity->setQueryFlags(mask);
	}
	if(m_dimEntity != NULL)
	{
		m_dimEntity->setQueryFlags(mask);
	}
}

void Boid::reset(void)
{
	m_position = m_initialPosition;
	m_velocity = m_initialVelocity;
	m_oldVelocity.set(0,0,0);
	//m_acceleration.set(0,0,0);
	m_flockcenter.set(0,0,0);
	m_active = false;
	SimObject::update(m_lastUpdate, false); 
}

Vector3 Boid::getFlockCenter(void)
{
	return m_flockcenter;
}

bool Boid::update(const double &elapsedSeconds, bool renderOnly) 
{
	bool ret =  true;
	if((m_boidType == PREY) || (m_boidType == PREDATOR))
	{
		m_position.y = 3;
		SimObject::update(elapsedSeconds, false);
		m_position.y = 0.0;
		return ret;
	}

	if(m_active)
	{
		double dt = elapsedSeconds - m_lastUpdate;

		m_position += (m_oldVelocity*dt + 0.5*m_acceleration*dt*dt);

		m_velocity += (m_acceleration*dt);

		// keep maximum allowed velocity
		if (m_velocity.length() > m_maxVelocity) 
		{
			m_velocity.setMagnitude(m_maxVelocity);
		}

		m_yaw = atan2(m_oldVelocity.x, m_oldVelocity.z);

		m_oldVelocity = m_velocity; 

		m_acceleration = navigator(renderOnly); 

		ret &= SimObject::update(elapsedSeconds, renderOnly); 

		if((m_position.x < DESK_LEFT) || (m_position.x > DESK_RIGHT) || (m_position.z > DESK_BACK) || (m_position.x < DESK_FRONT))
		{
			reset();
			return false;
		}
	} else 
	{
		m_lastUpdate = elapsedSeconds;
	}

	return ret;
}

Vector3 Boid::navigator(bool renderOnly) 
{
	Vector3 vacc(0, 0, 0);  

	if (accumulate(vacc, collisionAvoidance()) >= 1.0)           vacc *= m_maxAcceleration * m_maxVelocity;
	if(renderOnly == false && m_boidType == NORMAL)
	{
		// calculate the visibility and closest neighbor
		calculateNeighbors();

		switch(m_methodType)
		{
		case BOIDS_MODE :
			if (accumulate(vacc, flockCentering()) >= 1.0)  vacc *= m_maxAcceleration * m_maxVelocity; 
			if (accumulate(vacc, wander()) >= 1.0)  vacc *= m_maxAcceleration * m_maxVelocity; 
			break;
		case TARGET_MODE :
			if (accumulate(vacc, flockCentering()) >= 1.0)  vacc *= m_maxAcceleration * m_maxVelocity; 
			break;
		case AVOID_MODE :
			if (accumulate(vacc, flockCentering()) >= 1.0)  vacc *= m_maxAcceleration * m_maxVelocity; 
			if (accumulate(vacc, maintainingCruisingDistance()) >= 1.0)  vacc *= m_maxAcceleration * m_maxVelocity; 
			break;
		case COORDINATE_MODE :
			if (accumulate(vacc, flockCentering()) >= 1.0)  vacc *= m_maxAcceleration * m_maxVelocity; 
			if (accumulate(vacc, maintainingCruisingDistance()) >= 1.0)  vacc *= m_maxAcceleration * m_maxVelocity; 
			if (accumulate(vacc, velocityMatching()) >= 1.0)  vacc *= m_maxAcceleration * m_maxVelocity; 
			break;
		}
		if (accumulate(vacc, predatorAvoidance()) >= 1.0)  vacc *= m_maxAcceleration * m_maxVelocity; 
	}

	vacc.y = 0.0f;

	return(vacc);
}
  
void Boid::calculateNeighbors(void) 
{	
	Boid *n;   
	int val;
	Vector3 tempVec;
	double tempDistance;

	// initialized
	m_boids_observed = 0;			
	m_flockcenter.set(0,0,0);		  
	m_distanceToClosestNeighbor = MAX_NUM;
	m_foundClosestNeighbor = 0;
	m_closestNeighbor = NULL;

	// foreach boid, if it's not visible, put -1 in the visible matrix
	vector<Boid*>::iterator iter;
	for(iter = m_boidList.begin(); iter != m_boidList.end(); iter++) {
		n = (Boid*) *iter;
		if(n == NULL) continue;
		if (n->m_drawMode == ALL_HIDE_MODE) break;

		if (n == this) { // itself
			val = -1;
		} else {
			val = (visibleToSelf(n) ? n->m_boidType : -1);
		}
		setVisMatrix(n, val);

		if(val == -1) continue;
		if((val != m_boidType) && m_flockSelectively) continue; 

		m_flockcenter += n->m_position;
		m_boids_observed++;

		// distance 
		tempVec = n->m_position - m_position;
		tempDistance = tempVec.length();
		if (tempDistance < m_distanceToClosestNeighbor) 
		{
			if((m_lastUpdate < n->m_lastUpdate) && (n->getClosestNeightbor() == m_id))
			{
			} else
			{
				m_distanceToClosestNeighbor = tempDistance;
				m_foundClosestNeighbor = 1;
				m_closestNeighbor = n;
			}
		}
	}
}

double Boid::accumulate(Vector3 &accumulator, Vector3 valueToAdd) 
{
	double magnitudeRemaining = 1.0 - accumulator.length();
	double vMag = valueToAdd.length();
	double newMag = vMag < magnitudeRemaining ? vMag : magnitudeRemaining;

	accumulator += (direction(valueToAdd) * newMag); 

	return newMag;
}

Vector3 Boid::collisionAvoidance(void) 
{
	Obstacle* obs;
	ISectData d;
	Vector3 normalToObject(0, 0, 0);
	int objectSeen = 0;
	Vector3 pointOnObject;

	// ignore obstacles that are out of the range of our probe.
	double distanceToObject = getProbeLength() * 2;
	Vector3 tempVec;
	double calculatedDistance;
	
	// find closest collision with obstacle
	vector<Obstacle*>::iterator iter;
	for(iter = Obstacle::m_obstacleList.begin(); iter != Obstacle::m_obstacleList.end(); iter++) {
		obs = (Obstacle*) *iter;
		
		d = obs->doesRayIntersect(direction(m_velocity), m_position);
		if (d.m_intersectionflag == 1) {      
			tempVec = d.m_point-m_position;
			calculatedDistance = tempVec.length();
			if (calculatedDistance <= distanceToObject) {
				// found a closer obstacle
				obs->addEnergy();
				objectSeen = 1; 
				distanceToObject = calculatedDistance;
				normalToObject = d.m_normal;
				pointOnObject = d.m_point; 
				break;
			}
		}
	}

	if (!objectSeen) {
		m_dimEntity->setMaterialName("yellow_dark");
		return Vector3(0,0,0);
	}

	//-----------------------------------------------------------------------
	// resolve collision
	//-----------------------------------------------------------------------
	// the collided boid wants to head in the direction of the normal
	// so, push it away from the obstacle if the normal is pointing towards the boid.
	m_dimEntity->setMaterialName("red_dark");

	tempVec = -tempVec;

	if((tempVec.x > 0 && normalToObject.x >= 0) || (tempVec.x < 0 && normalToObject.x <= 0))
	{
		if((tempVec.z > 0 && normalToObject.z >= 0) || (tempVec.z < 0 && normalToObject.z <= 0)) 
		{
			if (m_velocity.angleBetween(normalToObject) < M_PI/2) 
				normalToObject =- normalToObject;
		}
	}
	
	double mag = 2- distanceToObject/getProbeLength();

	// ignore objects that are farther away than our probe.
	if (mag < 0) mag = 0;

	return direction(normalToObject) * mag;
}

Vector3 Boid::flockCentering(void) 
{
	Vector3 temp;

	if(m_target != NULL)
	{
		// now calculate a vector to head towards center of flock
		temp = m_target->m_position - m_position;

		// and the percentage of maximum acceleration (in decimal) to use when yaw toward center
		temp.setMagnitude(1.0f); 

	} else {
		if (m_boids_observed != 0) { 
			m_flockcenter /= m_boids_observed;
		 
			// now calculate a vector to head towards center of flock
			temp = m_flockcenter - m_position;

			// and the percentage of maximum acceleration (in decimal) to use when yaw toward center
			temp.setMagnitude(1.0f); 
		}
		else {
			// couldn't see any other birds
			temp.setMagnitude(0);
		}
	}

	return(temp);
}

Vector3 Boid::predatorAvoidance(void) 
{
	Vector3 temp(0, 0, 0);
	int objectSeen = 0;
	double distance = getProbeLength();
	double rangeDistance = distance * 10;

	if (m_splitter != NULL) 
	{
		Vector3 normalToObject;
		Vector3 pointOnObject;

		Vector3 radiusVector = m_position - m_splitter->m_position;
		double distanceToObject = radiusVector.length();
		if(distanceToObject < rangeDistance)
		{
			objectSeen = 1; 
			normalToObject = radiusVector;
			pointOnObject = m_splitter->m_position; 
		}
		if(objectSeen == 1)
		{
			normalToObject = -normalToObject; 

			//if (m_velocity.angleBetween(normalToObject) < M_PI/2) 
			//	normalToObject =- normalToObject;

			double mag = 8 - distanceToObject/distance;

			// ignore objects that are farther away than our probe.
			if (mag < 0) mag = 0;

			return direction(normalToObject) * mag * 100;
		}
	}

	return (temp);
}

Vector3 Boid::maintainingCruisingDistance(void) 
{
	vector<Boid*>::iterator iter;
	Vector3 speedAdjustmentVector(0, 0, 0);  

	if (m_foundClosestNeighbor) {
		// try to keep cruise distance as possble as they can
		// if it's close then stay away, if it's too far, stay closer.
		Vector3 separationVector = m_closestNeighbor->getPosition() - m_position;

		float separateFactor = 0.9;
		float approachFactor = 0.5;


		if(separationVector.length() < m_cruiseDistance)
		{
			if(separationVector.x > 0)
			{
				speedAdjustmentVector.x += separateFactor;
			} else 
			{
				speedAdjustmentVector.x -= separateFactor;
			}
			if(separationVector.z > 0)
			{
				speedAdjustmentVector.z += separateFactor;
			} else 
			{
				speedAdjustmentVector.z -= separateFactor;
			}
		} else if(separationVector.length() > m_cruiseDistance)
		{
			if(separationVector.x > 0)
			{
				speedAdjustmentVector.x -= approachFactor;
			} else 
			{
				speedAdjustmentVector.x += approachFactor;
			}
			if(separationVector.z > 0)
			{
				speedAdjustmentVector.z -= approachFactor;
			} else 
			{
				speedAdjustmentVector.z += approachFactor;
			}
		}
    
	}

	return(speedAdjustmentVector);
}

Vector3 Boid::velocityMatching(void) 
{
	Vector3 velocityOfClosestNeighbor(0,0,0);

	// If there is a close boid, set the percentage of our acceleration
	// --> for flying parallel 
	if (m_distanceToClosestNeighbor != MAX_NUM) {
		velocityOfClosestNeighbor = m_closestNeighbor->m_velocity;
		// with the velocity vector of closest boid, try to match it
		velocityOfClosestNeighbor.setMagnitude(0.2);
	}

	return(velocityOfClosestNeighbor);
}

Vector3 Boid::wander(void) 
{
	double distanceFromCruiseSpeed = (m_velocity.length() - (0.2 * m_maxVelocity)) / m_maxVelocity;

	double urgency = fabs(distanceFromCruiseSpeed);
	if (urgency > 0.15) urgency = 0.15;
  
	return direction(m_velocity) * urgency * (distanceFromCruiseSpeed > 0 ? -1 : 1);
}

void Boid::calculateRollPitchYaw(Vector3 appliedAcceleration, Vector3 currentVelocity, Vector3 currentPosition) 
{
	Vector3 lateralDir = direction((currentVelocity.crossProduct(appliedAcceleration)).crossProduct(currentVelocity));

	double lateralMag = appliedAcceleration.dotProduct(lateralDir);

	if (lateralMag == 0)
		m_roll = 0;  
	else
		m_roll = -atan2(getGravAcceleration(), lateralMag) + M_PI/2;

	m_pitch = -atan(currentVelocity.y / sqrt(currentVelocity.z*currentVelocity.z + currentVelocity.x*currentVelocity.x));
	
	m_yaw = atan2(currentVelocity.x, currentVelocity.z);

}

int Boid::visibleToSelf(Boid *b)
{
	Vector3 vectorToObject = b->m_position - m_position;
	double length = vectorToObject.length();

	int ret = (length < 200);
	// if it's not in range, check FOV
	//if ((ret <= 0) && (length < 500))
	if (ret <= 0) 
	{
		ret = (m_velocity.angleBetween(vectorToObject) <= 1.0471967);
	}

	return ret;
}

void Boid::setTarget(Boid* target)
{
	m_target = target;
}

void Boid::setSplitter(Boid* target)
{
	m_splitter = target;
}

int Boid::getClosestNeightbor(void)
{
	if(m_closestNeighbor) 
		return m_closestNeighbor->getId();
	return -1;
}

void Boid::setInitialVectors(Ogre::Vector3 pos, Ogre::Vector3  dir)
{
	m_position = pos;
	m_initialPosition= pos;
	m_velocity =  dir;
	m_initialVelocity= dir;
	SimObject::update(m_lastUpdate, false); 
}

void Boid::setFlockSelectively(bool flocking)
{
	m_flockSelectively = flocking;
}

void Boid::setMethod(methodTypes type)
{
	m_methodType = type; 
}

int Boid::checkVisMatrix(Boid* n)
{
	int j = n->getId();
	return m_visibilityMatrix[m_id][j];
}

void Boid::setVisMatrix(Boid* n,  int flag)
{
	int j = n->getId();
	m_visibilityMatrix[m_id][j] = flag;
}

double Boid::getGravAcceleration(void) 
{
	return 9.806650;
}

int Boid::getType(void) 
{
	return m_boidType;
}

float Boid::getProbeLength(void)
{
	float maxScale = 5;

	// if it's already maxVelocity, scalefactor = maxScale.
	// if velocity is 0, scalefactor = 1.
	float scaleFactor = ((maxScale-1)/m_maxVelocity) * m_velocity.length() + 1;

	return m_bodyLength * scaleFactor;
}

double Boid::setSpeed(void) 
{
  return 0.2 * m_maxVelocity;
}

void Boid::setMaxAcceleration(double acc)
{
	m_maxAcceleration = acc;
}

double Boid::setCruiseDistance(double distance)
{
	/*m_cruiseDistance = 2 * m_dimensions.z;
	if(m_cruiseDistance < distance)
		m_cruiseDistance = distance;*/
	m_cruiseDistance = distance;
	return m_cruiseDistance;
}

void Boid::setMaxVelocity(double velocity)
{
	m_maxVelocity = velocity;
}

double Boid::getMaxVelocity(void)
{
	return m_maxVelocity;
}

void Boid::initThread(void)
{
	int size = m_boidList.size();
	int threadcount = size / 20;
	if((size % 20) > 0) threadcount++;
	
	threadcount = 1;
	m_start = true;
	BoidThreads* boidThread = NULL;
	for(int i=0; i <threadcount; i++)
	{
		pthread_t thId;
		boidThread = new BoidThreads();
		boidThread->mId = i;
		boidThread->mBoid = this;
		if (pthread_create(&thId, 0, LoopThread, (void*)boidThread) != 0) 
		{
				//cout << getID() << " :  HSPEventHandler THREAD ERROR!!!" << endl;
		}
		boidThread->mthId = thId;
		m_thIdList.push_back(boidThread);
	}

}

void Boid::destroyThread(void)
{
	m_start = false;
	int status;
	vector<BoidThreads*>::iterator iter;
	BoidThreads* boidthreadptr;
	for(iter = m_thIdList.begin(); iter != m_thIdList.end(); iter++)
	{
		boidthreadptr = (BoidThreads*) *iter;
		pthread_join(boidthreadptr->mthId, (void **)&status);
		delete boidthreadptr;
		boidthreadptr = NULL;
	}
	m_thIdList.clear();
}

bool Boid::isRun(void)
{
	return m_start;
}

void* Boid::LoopThread(void *args) 
{
	BoidThreads* boidThread = (BoidThreads*)args;
	int portionID = 0;
	Boid* boidptr = NULL;
	int index;
	double mElapsedTime = 0;
	do
	{
		for(int i= 0; i < 100; i++)
		{
			index = (portionID*20) + i;
			boidptr = (Boid*) Boid::m_boidList[index];
			if(boidptr->m_drawMode == ALL_HIDE_MODE) break;
			boidptr->update(mElapsedTime, false);
		}

		Sleep(10);
		mElapsedTime += 1.0/UPDATEFPS;

	} while(boidThread->mBoid->isRun());
	
	pthread_exit(NULL);
	return 0;
}
