// Instancing.cpp : 
// Animation Class CS527
// Hyejung Hur

#include "Instancing.h"
#include "BoidTypes.h"


#ifdef __cplusplus
extern "C" {
#endif

#if OGRE_PLATFORM == OGRE_PLATFORM_WIN32
#define WIN32_LEAN_AND_MEAN
#include "windows.h"

	INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR strCmdLine, INT )
#else
	int main(int argc, char **argv)
#endif
	{
		// Create application object
		InstancingApplication app;

		try {
			app.go();
		} catch( Ogre::Exception& e ) {
#if OGRE_PLATFORM == OGRE_PLATFORM_WIN32
			MessageBox( NULL, e.getFullDescription().c_str(), "An exception has occured!", MB_OK | MB_ICONERROR | MB_TASKMODAL );
#else
			std::cerr << "An exception has occured: " << e.getFullDescription();
#endif
		}


		return 0;
	}
#ifdef __cplusplus
}
#endif


InstancingListener::InstancingListener(RenderWindow* win, Camera* cam,CEGUI::Renderer* renderer, InstancingApplication*main)
: ExampleFrameListener(win, cam,false,true),
mRequestShutDown(false), mLMBDown(false), mRMBDown(false), mAvgFrameTime(0.1), mlastUpdate(0.0f),
mMain(main), mNUMBOIDS(0), mDrawMode(0), mBoidsRate(0.0f), mBugNodeptr(NULL), mpSoundDevice(NULL),
mRefreshCount(0), mElapsedTime(0.0f), mKeyPressed(false), mStartStatus(false), mSoundPlay(false), 
m_preyptr(NULL), m_predatorptr(NULL), m_showDimension(SimObject::NORMAL_MODE), mGrabbedEntityObj(NULL)
{ 
	mSceneMgr = mMain->getSeceneManager();
	const GpuProgramManager::SyntaxCodes &syntaxCodes = GpuProgramManager::getSingleton().getSupportedSyntax();
	for (GpuProgramManager::SyntaxCodes::const_iterator iter = syntaxCodes.begin();iter != syntaxCodes.end();++iter)
	{
		LogManager::getSingleton().logMessage("supported syntax : "+(*iter));
	}
	mGUIRenderer=renderer;
	mMoveSpeed = 100.0f;
	mGrabbedEntity = -1;
	mSkipFrames = 2;

	mMouse->setEventCallback(this);

	mDebugOverlay->hide(); 
	
	createBoids();
	createBoundBox();
	mBoidsReadyList.clear();

    // create RaySceneQuery
    mRaySceneQuery = mSceneMgr->createRayQuery(Ray());
	//mRaySceneQuery->setWorldFragmentType(SimObject::WORLD_MASK);

	mpSoundDevice = alcOpenDevice("DirectSound3D");
	mpSoundContext=alcCreateContext(mpSoundDevice,NULL);
	alcMakeContextCurrent(mpSoundContext);
	alGetError();
	loadSound("background.wav", true);
}

//-----------------------------------------------------------------------
InstancingListener::~InstancingListener()
{
	int size = mRangeBoxList.size();
	for (int i=0; i < size; i++) 
	{
		delete mRangeBoxList[i];
	}
	mRangeBoxList.clear();

	for (int i=0; i < MAXBOIDS; i++)
	{
		delete mBoidsList[i];
	}	
	mBoidsList.clear();
	if(m_preyptr)
		delete m_preyptr;
	if(m_predatorptr)
		delete m_predatorptr;

	alDeleteSources(1,&mALaSource);
	alDeleteBuffers(1,&mALSampleSet);

	mpSoundContext=alcGetCurrentContext();
	mpSoundDevice=alcGetContextsDevice(mpSoundContext);
	alcMakeContextCurrent(NULL);
	alcDestroyContext(mpSoundContext);
	alcCloseDevice(mpSoundDevice);
}


//-----------------------------------------------------------------------
bool InstancingListener::frameStarted(const FrameEvent& evt)
{
	if(mRequestShutDown) return false;

	if(mStartStatus == true)
	{
		if(mRefreshCount == mSkipFrames)
		{
			updateBoids(false);
			mRefreshCount = 0;
		} else {
			updateBoids(true);
			mRefreshCount++;
		}
	}

	const bool returnValue = ExampleFrameListener::frameStarted(evt);


	return returnValue;
}

void InstancingListener::start(void)
{
	/*if(mBoidsList.size() > 0)
	{
		mBoidsList[0]->initThread();
	}*/
	mStartStatus = true;
}
void InstancingListener::pause(void)
{
	/*if(mBoidsList.size() > 0)
	{
		mBoidsList[0]->destroyThread();
	}*/
	mStartStatus = false;
}

bool InstancingListener::processUnbufferedKeyInput(const FrameEvent& evt)
{
	bool ret = true;

	if(mKeyboard->isKeyDown(OIS::KC_1) && mTimeUntilNextToggle <=0)
	{
		if(mCamMode < 0)
		{
			Vector3 pos = m_preyptr->getPosition();
			mCamera->setPosition(pos.x, 500, pos.z);
			mCamera->pitch(Ogre::Degree(-90));
			mTimeUntilNextToggle = 0.5;
			mCamMode = 1;
		} else 
		{
			Vector3 pos = m_preyptr->getPosition();
			mCamera->setPosition(pos.x, 10, pos.z + 400); 
			mCamera->lookAt(pos.x, 0, 0);
			mTimeUntilNextToggle = 0.5;
			mCamMode = -1;
		}
	} else if(mKeyboard->isKeyDown(OIS::KC_2) && mTimeUntilNextToggle <=0)
	{
		if(mCamMode < 0)
		{
			Vector3 pos = m_predatorptr->getPosition();
			mCamera->setPosition(pos.x, 500, pos.z);
			mCamera->pitch(Ogre::Degree(-90));
			mTimeUntilNextToggle = 0.5;
			mCamMode = 1;
		} else 
		{
			Vector3 pos = m_predatorptr->getPosition();
			mCamera->setPosition(pos.x, 10, pos.z + 400);
			mCamera->lookAt(pos.x, 0, 0);
			mTimeUntilNextToggle = 0.5;
			mCamMode = -1;
		}

	} else if(mKeyboard->isKeyDown(OIS::KC_F1) && mTimeUntilNextToggle <=0)
	{
		if(m_showDimension == SimObject::NORMAL_MODE)
		{
			m_showDimension = SimObject::DIMENSION_MODE;
		} else 
		{
			m_showDimension = SimObject::NORMAL_MODE;
		}
		std::vector<Boid*>::iterator iter;
		Boid* boidptr;
		int index =0;
		for(iter = mBoidsList.begin(); iter != mBoidsList.end(); iter++, index++)
		{
			boidptr = (Boid*) *iter;
			if(index < mNUMBOIDS)
			{
				boidptr->setDimensionDrawMode(m_showDimension);  // DIMENSION_MODE
			} else 
			{
				boidptr->setDimensionDrawMode(SimObject::ALL_HIDE_MODE);
			}

		}
		if(m_preyptr)
			m_preyptr->setDimensionDrawMode(m_showDimension);
		if(m_predatorptr)
			m_predatorptr->setDimensionDrawMode(m_showDimension);
		for(int i = 0; i < mRangeBoxList.size(); i++)
		{
			mRangeBoxList[i]->setDimensionDrawMode(m_showDimension);
		}
		mTimeUntilNextToggle = 1;
	} else 
	{
		ret = ExampleFrameListener::processUnbufferedKeyInput(evt);
	}
	return ret;
}

void InstancingListener::setBoidsCount(int count)
{
	mBoidsReadyList.clear();
	std::vector<Boid*>::iterator iter;
	Boid* boidptr;
	int index =0;
	double lastTime = mBoidsList[0]->getLastUpdate();
	for(iter = mBoidsList.begin(); iter != mBoidsList.end(); iter++, index++)
	{
		boidptr = (Boid*) *iter;
		if(index < count)
		{
			boidptr->setDimensionDrawMode(SimObject::NORMAL_MODE);  // DIMENSION_MODE
			if(boidptr->isActive() == false)
			{
				mBoidsReadyList.push_back(boidptr);
			}
		} else
		{
			if(index < mNUMBOIDS)
			{
				boidptr->reset();
			}
			boidptr->setDimensionDrawMode(SimObject::ALL_HIDE_MODE);
		}

	}

	mNUMBOIDS = count;

	int mod = count / 15;
	mSkipFrames = mod;
	if(mSkipFrames < 2) mSkipFrames = 2;

}

void InstancingListener::setBoidsCount(int count, Vector3 pos, Vector3 attitude)
{
	mBoidsReadyList.clear();

	mNUMBOIDS = count;
	std::vector<Boid*>::iterator iter;
	Boid* boidptr;
	int index =0;
	int RAD = 10;
	Vector3 p, v;
	double mv;
	Vector3 origin(100, 0.0, 0.0);
	double lastTime = mBoidsList[0]->getLastUpdate();
	for(iter = mBoidsList.begin(); iter != mBoidsList.end(); iter++, index++)
	{
		boidptr = (Boid*) *iter;
		p.set(pos.x - 0.5*RAD - RAD*0.025, -68, pos.z - 0.5*RAD - RAD*0.025);
		//attitude.set(attitude.x - 0.5*RAD - RAD*0.025, 0.0f, attitude.z - 0.5*RAD - RAD*0.025);
		mv = boidptr->getMaxVelocity();
		v  = direction(attitude) * (mv)/4.0;	
		//boidptr->setPosition(p);
		//boidptr->setVelocity(v);
		boidptr->setInitialVectors(p, v);
		//boidptr->resetTime(lastTime);
		if(index < mNUMBOIDS)
		{
			boidptr->setDimensionDrawMode(SimObject::NORMAL_MODE);  // DIMENSION_MODE
			mBoidsReadyList.push_back(boidptr);
		} else 
		{
			boidptr->setDimensionDrawMode(SimObject::ALL_HIDE_MODE);
		}
		boidptr->setMaxAcceleration(5.5);
		boidptr->update(0.0f, true);
	}

	int mod = count / 15;
	mSkipFrames = mod;
	if(mSkipFrames < 2) mSkipFrames = 2;

}

int InstancingListener::createBoids(void)
{
	Vector3 p(0, 0, 0);
	Vector3 v(0, 0, 0);   // velocity vector
	Vector3 d(10, 2, 10); // dimensions of boid (RAD, height, length), 1, .2, .75

	char name[10];
	Boid* boidptr = NULL;

	for (int i=0; i < MAXBOIDS; i++) {
		// set up attitude, position and velocity.
		sprintf(name, "boid%d", i);
		if ((i % 2) == 0) 
		{
			boidptr = new Boid(p, v, d, mSceneMgr, name, "small_cube.mesh"); // cube.mesh razor
		} else 
		{
			boidptr = new Boid(p, v, d, mSceneMgr, name, "small_cylinder.mesh"); // cube.mesh razor
		}

		boidptr->setDimensionDrawMode(SimObject::ALL_HIDE_MODE);
		boidptr->update(0.0f, true);
		mBoidsList.push_back(boidptr);
	}

	return 1;
}

void InstancingListener::createBoundBox(void)
{
	ObstacleBox* obstacleBoxptr;
	// make bounding box
	Vector3 topLeftBackCorner(DESK_LEFT, 50, DESK_BACK);
	Vector3 bottomRightFrontCorner(DESK_LEFT+10, -130, DESK_FRONT);
	obstacleBoxptr = new ObstacleBox(topLeftBackCorner, bottomRightFrontCorner, mSceneMgr, "rangeboxL") ;
	obstacleBoxptr->setPosition(topLeftBackCorner);
	mRangeBoxList.push_back(obstacleBoxptr);

	topLeftBackCorner.set(DESK_RIGHT, 50, DESK_BACK);
	bottomRightFrontCorner.set(DESK_RIGHT+10, -130, DESK_FRONT);
	obstacleBoxptr = new ObstacleBox(topLeftBackCorner, bottomRightFrontCorner, mSceneMgr, "rangeboxR") ;
	obstacleBoxptr->setPosition(topLeftBackCorner);
	mRangeBoxList.push_back(obstacleBoxptr);

	topLeftBackCorner.set(DESK_LEFT, 50, DESK_FRONT+10);
	bottomRightFrontCorner.set(DESK_RIGHT, -130, DESK_FRONT);
	obstacleBoxptr = new ObstacleBox(topLeftBackCorner, bottomRightFrontCorner, mSceneMgr, "rangeboxF") ;
	obstacleBoxptr->setPosition(topLeftBackCorner);
	mRangeBoxList.push_back(obstacleBoxptr);

	topLeftBackCorner.set(DESK_LEFT, 50, DESK_BACK+10);
	bottomRightFrontCorner.set(DESK_RIGHT, -130, DESK_BACK);
	obstacleBoxptr = new ObstacleBox(topLeftBackCorner, bottomRightFrontCorner, mSceneMgr, "rangeboxB") ;
	obstacleBoxptr->setPosition(topLeftBackCorner);
	mRangeBoxList.push_back(obstacleBoxptr);

}

void InstancingListener::setMaxVelocity( double acc )
{
	std::vector<Boid*>::iterator iter;
	Boid* boidptr;
	for(iter = mBoidsList.begin(); iter != mBoidsList.end(); iter++)
	{
		boidptr = (Boid*) *iter;
		boidptr->setMaxVelocity(acc); //setMaxAcceleration
	}
}

double InstancingListener::setCruiseDistance( double distance )
{
	std::vector<Boid*>::iterator iter;
	Boid* boidptr;
	double ret = 0.0;
	for(iter = mBoidsList.begin(); iter != mBoidsList.end(); iter++)
	{
		boidptr = (Boid*) *iter;
		ret = boidptr->setCruiseDistance(distance); //setMaxAcceleration
	}
	return ret;
}

void InstancingListener::setMethod(int type)
{
	if(type == 0) 
	{
		std::vector<Boid*>::iterator iter;
		Boid* boidptr;
		for(iter = mBoidsList.begin(); iter != mBoidsList.end(); iter++)
		{
			boidptr = (Boid*) *iter;
			boidptr->setTarget(NULL); 
		}
		Boid::setMethod(Boid::BOIDS_MODE);
	} else if(type == 1)
	{
		std::vector<Boid*>::iterator iter;
		Boid* boidptr;
		for(iter = mBoidsList.begin(); iter != mBoidsList.end(); iter++)
		{
			boidptr = (Boid*) *iter;
			boidptr->setTarget(NULL); 
		}
		Boid::setMethod(Boid::TARGET_MODE);
	} else if(type == 2)
	{
		std::vector<Boid*>::iterator iter;
		Boid* boidptr;
		for(iter = mBoidsList.begin(); iter != mBoidsList.end(); iter++)
		{
			boidptr = (Boid*) *iter;
			boidptr->setTarget(m_preyptr); 
		}
		Boid::setMethod(Boid::TARGET_MODE);
	} else if(type == 3)
	{	
		Boid::setMethod(Boid::AVOID_MODE);
	} else if(type == 4)
	{
		Boid::setMethod(Boid::COORDINATE_MODE);
	}
}

void InstancingListener::updateBoids(bool renderOnly)
{
	mlastUpdate += 1.0/UPDATEFPS;

	std::vector<Boid*>::iterator iter;
	Boid* boidptr;
	if (mlastUpdate >= mBoidsRate)
	{
		mlastUpdate = 0.0f;
		if(mBoidsReadyList.size() > 0)
		{
			iter = mBoidsReadyList.begin();
			boidptr = (Boid*) *iter;
			boidptr->setActive(true);
			mBoidsReadyList.erase(iter);
		}
	}

	bool ret;
	int index =0;
	for(iter = mBoidsList.begin(); iter != mBoidsList.end(); iter++, index++)
	{
		boidptr = (Boid*) *iter;
		ret = boidptr->update(mElapsedTime, renderOnly);
		if(ret == false)
		{
			mBoidsReadyList.push_back(boidptr);
		}
	}

	ret = false;
	int size = mRangeBoxList.size();
	for (int i=4; i < size; i++) 
	{
		ret |= mRangeBoxList[i]->update(mElapsedTime, false);
	}
	if((ret == true) && (mSoundPlay == false))
	{
		alSourcePlay(mALaSource); 
		mSoundPlay = true;
	} else if((mSoundPlay == true) && (ret == false))
	{ 
		//alSourceStop(mALaSource);
		alSourcePause(mALaSource);
		mSoundPlay = false;
	}

	//m_preyptr->update(mElapsedTime, true);
	//m_predatorptr->update(mElapsedTime, true);
	// increment elapsed time
	mElapsedTime += 1.0/UPDATEFPS;
} 

//-----------------------------------------------------------------------
bool InstancingListener::mouseMoved ( const OIS::MouseEvent &arg )
{
	CEGUI::System::getSingleton().injectMouseMove( arg.state.X.rel, arg.state.Y.rel );

    if ((mRMBDown || mLMBDown) && (mGrabbedEntity >= 0))		// if we are dragging the left mouse button
    {
        CEGUI::Point mousePos = CEGUI::MouseCursor::getSingleton().getPosition();
        Ray mouseRay = mCamera->getCameraToViewportRay(mousePos.d_x/float(arg.state.width),mousePos.d_y/float(arg.state.height));
        mRaySceneQuery->setRay(mouseRay);
		mRaySceneQuery->setQueryMask(SimObject::WORLD_MASK | SimObject::MOVABLE_MASK); 

        RaySceneQueryResult &result = mRaySceneQuery->execute();
        RaySceneQueryResult::iterator itr = result.begin();
		Obstacle* obstracleptr = NULL;
		int size = result.size();
		for(int i =0; i < size; i++, itr++)
		{
			if (result[i].movable->getName() == "desktop") 
			{
				Vector3 newPos = mouseRay.getOrigin() + mouseRay.getDirection() * result[i].distance;
				newPos.y = -68;

				if(mGrabbedEntity == 0)
				{
					m_preyptr->setPosition(newPos);
					m_preyptr->update(0.0f, true);
				} else if(mGrabbedEntity == 1)
				{
					m_predatorptr->setPosition(newPos);
					m_predatorptr->update(0.0f, true);
				} else 
				{
					if(mGrabbedEntityObj)
					{
						mGrabbedEntityObj->setPosition(newPos);
						mGrabbedEntityObj->update(mElapsedTime, true);
					}
				} 
				break;
			} else 
			{
				Vector3 newPos = mouseRay.getOrigin() + mouseRay.getDirection() * result[i].distance;
				newPos.y = -68;

				if(mGrabbedEntity == 0)
				{
					m_preyptr->setPosition(newPos);
					m_preyptr->update(0.0f, true);
				} else if(mGrabbedEntity == 1)
				{
					m_predatorptr->setPosition(newPos);
					m_predatorptr->update(0.0f, true);
				}
				break;
			}
		}

    } 

	return true;
}	

//-----------------------------------------------------------------------
bool InstancingListener::mousePressed ( const OIS::MouseEvent &arg, OIS::MouseButtonID id )
{
	CEGUI::System::getSingleton().injectMouseButtonDown(convertOISMouseButtonToCegui(id));

    if (id == OIS::MB_Left)				// left mouse button down
    {
        onLeftPressed(arg);
        mLMBDown = true;
    } else if (id == OIS::MB_Right)		// right mouse button down
    {
		onLeftPressed(arg);
        mRMBDown = true;
    }

	return true;
}
//-----------------------------------------------------------------------
bool InstancingListener::mouseReleased (const OIS::MouseEvent &arg, OIS::MouseButtonID id)
{
	CEGUI::System::getSingleton().injectMouseButtonUp(convertOISMouseButtonToCegui(id));

	mLMBDown = false;
	mGrabbedEntity = -1;
	mRMBDown = false;
	mGrabbedEntityObj = NULL;
	return true;
}

// specific handlers
void InstancingListener::onLeftPressed(const OIS::MouseEvent &arg)
{
    // setup the ray scene query, use CEGUI's mouse position
    CEGUI::Point mousePos = CEGUI::MouseCursor::getSingleton().getPosition();
    Ray mouseRay = mCamera->getCameraToViewportRay(mousePos.d_x/float(arg.state.width), mousePos.d_y/float(arg.state.height));
    mRaySceneQuery->setRay(mouseRay);
	mRaySceneQuery->setQueryMask(SimObject::MOVABLE_MASK | SimObject::OBSTACLE_MASK); 

    // execute query
    RaySceneQueryResult &result = mRaySceneQuery->execute();
    RaySceneQueryResult::iterator itr = result.begin( );
	int size = result.size();
	std::string temp;
	for(int i =0; i < size; i++, itr++)
    {
		temp = result[i].movable->getName();
		 // "desktop"
		if ((result[i].movable->getName() == "prey-dim") || (result[i].movable->getName() == "prey"))
		{
			mGrabbedEntity = 0;
			break;
		} else if ((result[i].movable->getName() == "predator-dim") || (result[i].movable->getName() == "predator")) 
		{
			mGrabbedEntity = 1;
			break;
		} else if (temp.find("flower") !=  std::string::npos) 
		{
			mGrabbedEntityObj = getObstacle(temp.c_str());
			mGrabbedEntity = 2;
			break;
		} else if (temp.find("penguine") !=  std::string::npos) 
		{
			mGrabbedEntityObj = getObstacle(temp.c_str());
			mGrabbedEntity = 2;
			break;
		} else if (temp.find("note") !=  std::string::npos) 
		{
			mGrabbedEntityObj = getObstacle(temp.c_str());
			mGrabbedEntity = 2;
			break;
		} else if (temp.find("dog") !=  std::string::npos) 
		{
			mGrabbedEntityObj = getObstacle(temp.c_str());
			mGrabbedEntity = 2;
			break;
		}

    } 
}


Obstacle* InstancingListener::getObstacle(const char* name)
{
	Obstacle* obstacleptr = NULL;

	std::vector<Obstacle*>::iterator iter;
	for(iter = mRangeBoxList.begin(); iter != mRangeBoxList.end(); iter++)
	{
		obstacleptr = (Obstacle*) *iter;
		if(strcmp(obstacleptr->getName(),name) ==0)
		{
			break; 
		}
	}
	return obstacleptr;
}

void InstancingListener::requestShutdown(void)
{
	mRequestShutDown=true;
}

bool InstancingListener::handleMouseMove(const CEGUI::EventArgs& e)
{
	using namespace CEGUI;

	if( mLMBDown)
	{
		int a =0;
		// rotate camera
		mRotX += Ogre::Degree(-((const MouseEventArgs&)e).moveDelta.d_x * mAvgFrameTime * 10.0);
		mRotY += Ogre::Degree(-((const MouseEventArgs&)e).moveDelta.d_y * mAvgFrameTime * 10.0);
		mCamera->yaw(mRotX);
		mCamera->pitch(mRotY);
		MouseCursor::getSingleton().setPosition( mLastMousePosition );
	}


	return true;
}
//--------------------------------------------------------------------------
bool InstancingListener::handleMouseButtonUp(const CEGUI::EventArgs& e)
{
	using namespace CEGUI;


	if( ((const MouseEventArgs&)e).button == LeftButton )
	{
		mLMBDown = false;
		MouseCursor::getSingleton().setPosition( mLastMousePosition );
		CEGUI::MouseCursor::getSingleton().show();
	}


	return true;
}

//--------------------------------------------------------------------------
bool InstancingListener::handleMouseButtonDown(const CEGUI::EventArgs& e)
{
	using namespace CEGUI;


	if( ((const MouseEventArgs&)e).button == LeftButton )
	{
		mLMBDown = true;
		mLastMousePosition=CEGUI::MouseCursor::getSingleton().getPosition();
		CEGUI::MouseCursor::getSingleton().hide();
	}

	return true;
}

void InstancingListener::initalize(void)
{ 
	FILE *fptr= fopen("default.cfg","r+");
	if(fptr == NULL) return;
	char	line[MAX_LINE];
	memset(line, 0, MAX_LINE);

	char token[TOKEN_LEN];
	memset(token, 0, TOKEN_LEN);
	
	int nvalue;
	float fvalue[2];
	int mode = 0;
	char bName[10];
	Vector3 position(0, -70, 0);
	double fPosition[2];
	float maxVelocity, maxDistance;
	while(fgets(line, MAX_LINE, fptr)!=NULL)
	{
		mode = 0;
		getToken(line, token);
		toupper(&token[0]);
		if(strcmp("FLOWER", token) == 0)			mode = 1;
		else if(strcmp("PENGUIN", token) == 0)		mode = 2;
		else if(strcmp("NOTES", token) == 0)		mode = 3;
		else if(strcmp("PREDATOR", token) == 0)		mode = 4;
		else if(strcmp("PREY", token) == 0)			mode = 5;
		else if(strcmp("DOG", token) == 0)			mode = 6;
		else if(strcmp("MODE", token) == 0)
		{
			getToken(line, token);
			toupper(&token[0]);
			int mode = 0;
			if(strcmp("FLOCKING", token) == 0) mode = 1;
			else if(strcmp("TARGET", token) == 0) mode = 2;
			else if(strcmp("AVOID_DISTANCE", token) == 0) mode = 3;
			else if(strcmp("VELOCITY_MATCHING", token) == 0) mode = 4;
			mMain->updatePropertyPanel(mode);
		} else if(strcmp("BOIDS", token) == 0)	
		{
			getToken(line, token);
			nvalue = atoi(token);
			if(nvalue > 100) nvalue = 100;
			mMain->updatePropertyPanel(0, nvalue* 0.01f + 0.005f);
			getToken(line, token);
			toupper(&token[0]);
			if (mBugNodeptr == NULL)
			{
				Vector3 dir;
				mBugNodeptr  = mSceneMgr->getRootSceneNode()->createChildSceneNode();
				Entity* pnodeEnt = mSceneMgr->createEntity( "mug", "mug.mesh" );
				mBugNodeptr->attachObject( pnodeEnt );
				pnodeEnt->setQueryFlags(SimObject::MOVABLE_MASK);
				if(strcmp("LEFTTOP", token) == 0)
				{
					mBugNodeptr->setPosition(DESK_LEFT+50, -72, DESK_FRONT+50);
					dir.set(1, 0, 1);
					mBugNodeptr->yaw(Ogre::Degree(45));
				} else if(strcmp("LEFTBOTTOM", token) == 0)
				{
					mBugNodeptr->setPosition(DESK_LEFT+50, -72, DESK_BACK-50);
					mBugNodeptr->yaw(Ogre::Degree(135));
					dir.set(1, 0, -1);
				} else if(strcmp("RIGHTTOP", token) == 0)
				{
					mBugNodeptr->setPosition(DESK_RIGHT-50, -72, DESK_FRONT+50);
					mBugNodeptr->yaw(Ogre::Degree(-45));
					dir.set(-1, 0, 1);
				} else if(strcmp("RIGHTBOTTOM", token) == 0)
				{
					mBugNodeptr->setPosition(DESK_RIGHT-50, -72, DESK_BACK-50);
					mBugNodeptr->yaw(Ogre::Degree(-135));
					dir.set(-1, 0, -1);
				}
				setBoidsCount(nvalue, mBugNodeptr->getPosition(), dir);
			}
		} else if(strcmp("RATE", token) == 0)			
		{
			getToken(line, token);
			mBoidsRate = 6.0f / atof(token);
		} else if(strcmp("VELOCITY", token) == 0)
		{
			getToken(line, token);
			maxVelocity = atof(token);
			mMain->updatePropertyPanel(1, maxVelocity*0.1);
		} else if(strcmp("DISTANCE", token) == 0)
		{
			getToken(line, token);
			maxDistance = atof(token);
			mMain->updatePropertyPanel(2, maxDistance*0.1);
		}

		if(mode > 0)
		{
			// model no x, y
			getToken(line, token);
			nvalue = atoi(token);
			getToken(line, token);
			fvalue[0] = atof(token);
			getToken(line, token);
			fvalue[1] = atof(token);
			getXY(fvalue[0], fvalue[1], fPosition[0], fPosition[1]);
			position.set(fPosition[0], -70, fPosition[1]);
		}

		switch(mode)
		{
		case 1 :
			{
				sprintf(bName, "flower%d", nvalue);
				Vector3 dimension(40.214, 162.645, 63.723);
				ObstacleBox* obstacleBoxptr = new ObstacleBox(mSceneMgr, bName, position, dimension, "pot.mesh", "flower.mesh") ;
				obstacleBoxptr->setPosition(position);
				obstacleBoxptr->setDimensionDrawMode(SimObject::NORMAL_MODE);   // DIMENSION_MODE
				obstacleBoxptr->update(0.0f, true);
				mRangeBoxList.push_back(obstacleBoxptr);
			}
			break;
		case 2 :
			{
				sprintf(bName, "penguine%d", nvalue);
				Vector3 dimension(61.495, 111.289, 58.549);
				ObstacleBox* obstacleBoxptr = new ObstacleBox(mSceneMgr, bName, position, dimension, "foot.mesh", "penguine.mesh") ;
				obstacleBoxptr->setAniMaxAngle(15.0f);
				obstacleBoxptr->setPosition(position);
				obstacleBoxptr->setDimensionDrawMode(SimObject::NORMAL_MODE);   // DIMENSION_MODE
				obstacleBoxptr->update(0.0f, true);
				mRangeBoxList.push_back(obstacleBoxptr);
			}
			break;
		case 3 :
			{
				sprintf(bName, "note%d", nvalue);
				Vector3 dimension(69.22, 69.415, 47.131);
				ObstacleBox*  obstacleBoxptr = new ObstacleBox(mSceneMgr, bName, position, dimension, "notes.mesh", "heart.mesh") ;
				obstacleBoxptr->setPosition(position);
				obstacleBoxptr->setAniMaxAngle(12.0f);
				obstacleBoxptr->setDimensionDrawMode(SimObject::NORMAL_MODE);   // DIMENSION_MODE
				obstacleBoxptr->update(0.0f, true);
				mRangeBoxList.push_back(obstacleBoxptr);
			}
			break;
		case 4 :
			{
				if (m_predatorptr == NULL)
				{
					Vector3 dimension(50, 20, 50);
					position.y = -68;
					Vector3 v(0, 0, 0);
					// create predator - 102 Boid::PREDATOR
					m_predatorptr = new Boid(position, v, dimension, mSceneMgr, "predator", "red_predator.mesh", Boid::PREDATOR); 
					m_predatorptr->setDimensionDrawMode(SimObject::NORMAL_MODE);
					m_predatorptr->setActive(true);				
					m_predatorptr->update(0.0f, true);
					std::vector<Boid*>::iterator iter;
					Boid* boidptr;
					for(iter = mBoidsList.begin(); iter != mBoidsList.end(); iter++)
					{
						boidptr = (Boid*) *iter;
						boidptr->setSplitter(m_predatorptr);
					}
				}
			}
			break;
		case 5 :
			{
				if(m_preyptr == NULL)
				{
					// create prey - 101 Boid::PREY
					Vector3 dimension(50, 20, 50);
					position.y = -68;
					Vector3 v(0, 0, 0);
					m_preyptr = new Boid(position, v, dimension, mSceneMgr, "prey", "blue_prey.mesh", Boid::PREY); 
					m_preyptr->setDimensionDrawMode(SimObject::NORMAL_MODE);
					m_preyptr->setActive(true);
					m_preyptr->update(0.0f, true);
				}
			}
			break;
		case 6 :
			{
				sprintf(bName, "dog%d", nvalue);
				Vector3 dimension(97.943, 168.229, 107.436);
				ObstacleBox*  obstacleBoxptr = new ObstacleBox(mSceneMgr, bName, position, dimension, "dogbody.mesh", "doghead.mesh") ;
				obstacleBoxptr->setPosition(position);
				obstacleBoxptr->setAniMaxAngle(8.0f);
				obstacleBoxptr->setDimensionDrawMode(SimObject::NORMAL_MODE);   // DIMENSION_MODE
				obstacleBoxptr->update(0.0f, true);
				mRangeBoxList.push_back(obstacleBoxptr);
			}
			break;
		}

	}

	fclose(fptr);
	
}

void InstancingListener::getXY(double a, double b, double& x, double& y)
{
	// left/right : -840 ~ 880, front/back : 990 ~ -200
	double left = -840;
	double right = 880;
	double back = 560;
	double front = -330;

	x = DESK_LEFT + a * (DESK_RIGHT - DESK_LEFT);
	y = DESK_BACK - b * (DESK_BACK - DESK_FRONT);
}


void InstancingListener::loadSound(char* filename, bool looping)
{
	char*			alBuffer;
	ALenum			alFormatBuffer;
	ALsizei			alFreqBuffer;
	ALsizei			alBufferLen;

	ALboolean loop;
	alutLoadWAVFile(filename, &alFormatBuffer, (void **) &alBuffer,&alBufferLen, &alFreqBuffer, &loop);

	alGenSources(1, &mALaSource);
	alGenBuffers(1, &mALSampleSet);
	alBufferData(mALSampleSet, alFormatBuffer, alBuffer, alBufferLen, alFreqBuffer);
	alSourcei(mALaSource, AL_BUFFER, mALSampleSet);

	alutUnloadWAV(alFormatBuffer, alBuffer, alBufferLen, alFreqBuffer);			
	alSourcef(mALaSource,AL_PITCH,1.0f);
	alSourcef(mALaSource,AL_GAIN,1.0f);
	alSourcei(mALaSource,AL_LOOPING,AL_TRUE);

}
