// Obstacle.h : class for obstacles 
// Animation Class CS527
// Hyejung Hur

#include "Obstacle.h"
#include "BoidTypes.h"

using namespace std;
using namespace Ogre;

/* ########################################################################
 OBSTACLE
######################################################################### */
vector<Obstacle*>  Obstacle::m_obstacleList;
int Obstacle::m_obstracleCount =0;

Obstacle::Obstacle(void) 
: m_energy(0), m_angleStep(0.3f), m_enegyNode(NULL), m_enegyEntity(NULL), m_maxAngle(6.0f),
SimObject() 
{ 
	if(m_geomEntity != NULL)
	{
		m_geomEntity->setQueryFlags(OBSTACLE_MASK);
	}
}

Obstacle::Obstacle(SceneManager* mSceneMgr, char* name, char* filename, char* anifilename, double gab) 
: m_energy(0), m_angleStep(0.3f), m_enegyNode(NULL), m_enegyEntity(NULL), m_maxAngle(6.0f),
SimObject(mSceneMgr, name, filename, anifilename, gab) 
{ 
	if(m_geomEntity != NULL)
	{
		m_geomEntity->setQueryFlags(OBSTACLE_MASK);
	}
}
	
Obstacle::~Obstacle(void) 
{ 
	m_obstacleList.clear();
}

int Obstacle::getEnergy(void)
{
	return m_energy;
}	

void Obstacle::setAniMaxAngle(double  angle)
{
	m_maxAngle = angle;
}

void Obstacle::addEnergy(void)
{
	if(m_energy <= 2000)
		m_energy ++;
}
		
ISectData Obstacle::doesRayIntersect(Vector3 raydirection, Vector3 rayorigin) 
{
	return intersectionWithRay(raydirection, rayorigin);
}

void Obstacle::setPosition(Ogre::Vector3 newPosition)
{
	SimObject::setPosition(newPosition); 
}

bool Obstacle::update(const double &elapsedSeconds, bool renderOnly) 
{
	bool ret =  false;

	if(renderOnly == true)
	{
		SimObject::update(elapsedSeconds, renderOnly); 
	} else {
		double dt = elapsedSeconds - m_lastUpdate;
		if(m_aniGeoNode && (m_energy > 500))
		{
			m_angle += Ogre::Degree(m_angleStep);
			if ((m_angle.valueDegrees() < m_maxAngle) && (m_angle.valueDegrees() > -m_maxAngle))
			{
				m_aniGeoNode->roll(Ogre::Degree(m_angleStep));
			} else {
				m_angleStep *= -1;
			}
			m_energy -= 5;
			ret = true;
		} 
		m_energy -= 1;
		if (m_energy < 0) m_energy = 0;
		
		if(m_energy < 300)
		{
			m_enegyEntity->setMaterialName("red_dark");
		} else if(m_energy < 700)
		{
			m_enegyEntity->setMaterialName("yellow_dark");
		} else 
		{
			m_enegyEntity->setMaterialName("blue_color");
		}

	}
	return ret;
}


/* ########################################################################
 BOX
######################################################################### */
ObstacleBox::ObstacleBox(const Vector3 &topLeftBackCorner, const Vector3 &bottomRightFrontCorner, SceneManager* mSceneMgr, char* name, char* filename) 
: Obstacle(mSceneMgr, name, filename, NULL, 0.0f)
{
	if(filename == NULL && m_node == NULL)
	{
		m_node = mSceneMgr->getRootSceneNode()->createChildSceneNode();
		m_geomNode = m_node->createChildSceneNode();
		m_geomEntity= NULL;

		char dimname[15];
		sprintf(dimname, "%s-dim", name);
		m_dimNode = m_node->createChildSceneNode();
		m_dimEntity = mSceneMgr->createEntity( dimname, Ogre::SceneManager::PT_CUBE);

		AxisAlignedBox bound = m_dimEntity->getBoundingBox();
		Vector3 size = bound.getSize();
		m_dimensions.set(topLeftBackCorner.x - bottomRightFrontCorner.x, topLeftBackCorner.y - bottomRightFrontCorner.y, bottomRightFrontCorner.z - topLeftBackCorner.z);
		Vector3 scalefactor(m_dimensions.x/size.x, m_dimensions.y/size.y, m_dimensions.z/size.z);
		m_dimNode->setScale(scalefactor);
		//m_drawMode = DIMENSION_MODE;
		m_dimNode->setPosition(topLeftBackCorner.x + (bottomRightFrontCorner.x - topLeftBackCorner.x) / 2.0, bottomRightFrontCorner.y + (topLeftBackCorner.y - bottomRightFrontCorner.y) / 2.0, topLeftBackCorner.z + (bottomRightFrontCorner.z - topLeftBackCorner.z) / 2.0);
	}

	init(topLeftBackCorner, bottomRightFrontCorner);
}

ObstacleBox::ObstacleBox(const Vector3 &topLeftBackCorner, const Vector3 &bottomRightFrontCorner) 
: Obstacle()
{
	init(topLeftBackCorner, bottomRightFrontCorner);

}

ObstacleBox::ObstacleBox(const ObstacleBox &b, SceneManager* mSceneMgr, char* name, char* filename) 
: Obstacle(mSceneMgr, name, filename, NULL, 0.0f)
{
	for (int i= 0; i < 6; i++) {
		m_side[i] = new ObstaclePolygon(*(b.m_side[i]));
	}

	m_tlb = b.m_tlb;
	m_brf = b.m_brf;

	m_boundingSphere = new ObstacleSphere(*(b.m_boundingSphere));
}

ObstacleBox::ObstacleBox(SceneManager* mSceneMgr, char* name, Vector3 bPosition, Vector3 bDimensions, char* filename, char* anifilename, double gab)
: Obstacle(mSceneMgr, name, filename, anifilename, gab)
{
	m_position = bPosition;
	m_dimensions = bDimensions; 

	Vector3 topLeftBackCorner(bPosition.x - bDimensions.x/2, bPosition.y + bDimensions.y/2, bPosition.z - bDimensions.z/2);
	Vector3 bottomRightFrontCorner(bPosition.x + bDimensions.x/2, bPosition.y - bDimensions.y/2, bPosition.z + bDimensions.z/2);

	init(topLeftBackCorner, bottomRightFrontCorner);
	
	initScene(false);

	if(m_geomEntity != NULL)
	{
		char name[15];
		sprintf(name, "%d-energy", m_obstacleList.size());
		m_enegyNode = m_geomNode->createChildSceneNode();
		m_enegyEntity = mSceneMgr->createEntity( name, Ogre::SceneManager::PT_CUBE);
		m_enegyEntity->setMaterialName("red_dark");
		Ogre::AxisAlignedBox bound = m_geomEntity->getBoundingBox(); 
		Ogre::Vector3 size = bound.getSize();
		if(m_aniGeomEntity)
		{
			Ogre::AxisAlignedBox anibound = m_aniGeomEntity->getBoundingBox(); 
			Ogre::Vector3 anisize = anibound.getSize();
			size.y += anisize.y;
		}
		size.y += 20;
		m_enegyNode->attachObject( m_enegyEntity );
		m_enegyNode->setScale(Vector3(0.3, 0.1, 0.1));
		m_enegyNode->setPosition(0, size.y, 0);
	}

}

ObstacleBox::~ObstacleBox(void) 
{
	for (int i = 0; i < 6; i++)
		delete m_side[i];

	delete m_boundingSphere;
}

void ObstacleBox::setPosition(Ogre::Vector3 newPosition)
{
	m_position = newPosition;

	Vector3 topLeftBackCorner(m_position.x - m_dimensions.x/2, m_position.y + m_dimensions.y/2, m_position.z - m_dimensions.z/2);
	Vector3 bottomRightFrontCorner(m_position.x + m_dimensions.x/2, m_position.y - m_dimensions.y/2, m_position.z + m_dimensions.z/2);

	m_tlb = topLeftBackCorner;
	m_brf = bottomRightFrontCorner;
	double width  = m_tlb.x - m_brf.x;
	double height = m_tlb.y - m_brf.y;
	double length = m_tlb.z - m_brf.z;
	
	Ogre::Vector3 ** vList = NULL;
	for (int i = 0; i < 6; i++)
	{
		vList = m_side[i]->getVertex();
		if(vList == NULL) continue;
		update(i, vList, width, height, length);
		m_side[i]->update();
	}

	Vector3 center(m_brf.x + width/2, m_brf.y + height/2, m_brf.z + length/2);
	m_boundingSphere->update(center);
}

bool ObstacleBox::update(const double &elapsedSeconds, bool renderOnly)
{
	return Obstacle::update(elapsedSeconds, renderOnly);
}

void ObstacleBox::update(int idx, Ogre::Vector3 ** vList, double width, double height, double length)
{
	if(idx == 0) 
	{
		// Top
		vList[0]->set(m_brf.x + 0, m_brf.y + height, m_brf.z + 0);
		vList[1]->set(m_brf.x + 0, m_brf.y + height, m_brf.z + length);	
		vList[2]->set(m_brf.x + width, m_brf.y + height, m_brf.z + length);
		vList[3]->set(m_brf.x + width, m_brf.y + height, m_brf.z + 0);
	} else if(idx == 1) 
	{
		// Bottom
		vList[0]->set(m_brf.x, m_brf.y, m_brf.z);
		vList[1]->set(m_brf.x + 0, m_brf.y + height, m_brf.z + length);
		vList[2]->set(m_brf.x + width, m_brf.y + height, m_brf.z + length);
		vList[3]->set(m_brf.x + width, m_brf.y + height, m_brf.z + 0);
	} else if(idx == 2) 
	{
		// Left
		vList[0]->set(m_brf.x + width, m_brf.y, m_brf.z);
		vList[1]->set(m_brf.x + width, m_brf.y + height, m_brf.z + 0);
		vList[2]->set(m_tlb.x, m_tlb.y, m_tlb.z);
		vList[3]->set(m_brf.x + width, m_brf.y + 0, m_brf.z + length);
	} else if(idx == 3) 
	{
		// Right
		vList[0]->set(m_brf.x, m_brf.y, m_brf.z);
		vList[1]->set(m_brf.x, m_brf.y + height, m_brf.z);
		vList[2]->set(m_tlb.x - width, m_tlb.y, m_tlb.z);
		vList[3]->set(m_brf.x, m_brf.y, m_brf.z + length);	
	} else if(idx == 4) 
	{
		// Front
		vList[0]->set(m_brf.x, m_brf.y, m_brf.z);
		vList[1]->set(m_brf.x, m_brf.y + height, m_brf.z);
		vList[2]->set(m_brf.x + width, m_brf.y + height, m_brf.z);
		vList[3]->set(m_brf.x + width, m_brf.y, m_brf.z);		
	} else if(idx == 5) 
	{
		// Back
		vList[0]->set(m_brf.x, m_brf.y, m_brf.z + length);
		vList[1]->set(m_brf.x, m_brf.y + height, m_brf.z + length);
		vList[2]->set(m_brf.x + width, m_brf.y + height, m_brf.z + length);
		vList[3]->set(m_brf.x + width, m_brf.y, m_brf.z + length);	
	}
}

void ObstacleBox::init(const Vector3 &topLeftBackCorner, const Vector3 &bottomRightFrontCorner)
{
	m_id = m_obstracleCount;
	m_obstacleList.push_back(this);
	m_obstracleCount++;
	
	if(m_name.size() == 0)
	{
		char bName[10];
		sprintf(bName, "obstacle%d", m_id);
		m_name = bName;
	}
	cout << "tl : (" << topLeftBackCorner.x << " " << topLeftBackCorner.y << " " << topLeftBackCorner.z << ") ";
	cout << "br : (" << bottomRightFrontCorner.x << " " << bottomRightFrontCorner.y << " " << bottomRightFrontCorner.z << ") " << endl;
	

	m_tlb = topLeftBackCorner;
	m_brf = bottomRightFrontCorner;

	double width  = m_tlb.x - m_brf.x;
	double height = m_tlb.y - m_brf.y;
	double length = m_tlb.z - m_brf.z;

	// Top
	Vector3** vList = new Vector3*[4];
	vList[0] = new Vector3();
	vList[0]->set(m_brf.x + 0, m_brf.y + height, m_brf.z + 0);
	vList[1] = new Vector3();
	vList[1]->set(m_brf.x + 0, m_brf.y + height, m_brf.z + length);
	vList[2] = new Vector3();	
	vList[2]->set(m_brf.x + width, m_brf.y + height, m_brf.z + length);
	vList[3] = new Vector3();	
	vList[3]->set(m_brf.x + width, m_brf.y + height, m_brf.z + 0);
	m_side[0] = new ObstaclePolygon(4, vList, m_id);

	// Bottom
	vList = new Vector3*[4];
	vList[0] = new Vector3();
	vList[0]->set(m_brf.x, m_brf.y, m_brf.z);
	vList[1] = new Vector3();	
	vList[1]->set(m_brf.x + 0, m_brf.y + height, m_brf.z + length);
	vList[2] = new Vector3();	
	vList[2]->set(m_brf.x + width, m_brf.y + height, m_brf.z + length);
	vList[3] = new Vector3();	
	vList[3]->set(m_brf.x + width, m_brf.y + height, m_brf.z + 0);
	m_side[1] = new ObstaclePolygon(4, vList, m_id);

	// Left
	vList = new Vector3*[4];
	vList[0] = new Vector3();
	vList[0]->set(m_brf.x + width, m_brf.y, m_brf.z);
	vList[1] = new Vector3();	
	vList[1]->set(m_brf.x + width, m_brf.y + height, m_brf.z + 0);
	vList[2] = new Vector3();	
	vList[2]->set(m_tlb.x, m_tlb.y, m_tlb.z);
	vList[3] = new Vector3();	
	vList[3]->set(m_brf.x + width, m_brf.y + 0, m_brf.z + length);
	m_side[2] = new ObstaclePolygon(4, vList, m_id);

	// Right
	vList = new Vector3*[4];
	vList[0] = new Vector3();
	vList[0]->set(m_brf.x, m_brf.y, m_brf.z);
	vList[1] = new Vector3();	
	vList[1]->set(m_brf.x, m_brf.y + height, m_brf.z);
	vList[2] = new Vector3();	
	vList[2]->set(m_tlb.x - width, m_tlb.y, m_tlb.z);
	vList[3] = new Vector3();	
	vList[3]->set(m_brf.x, m_brf.y, m_brf.z + length);	
	m_side[3] = new ObstaclePolygon(4, vList, m_id);

	// Front
	vList = new Vector3*[4];
	vList[0] = new Vector3();
	vList[0]->set(m_brf.x, m_brf.y, m_brf.z);
	vList[1] = new Vector3();	
	vList[1]->set(m_brf.x, m_brf.y + height, m_brf.z);
	vList[2] = new Vector3();	
	vList[2]->set(m_brf.x + width, m_brf.y + height, m_brf.z);
	vList[3] = new Vector3();	
	vList[3]->set(m_brf.x + width, m_brf.y, m_brf.z);		
	m_side[4] = new ObstaclePolygon(4, vList, m_id);

	// Back
	vList = new Vector3*[4];
	vList[0] = new Vector3();
	vList[0]->set(m_brf.x, m_brf.y, m_brf.z + length);
	vList[1] = new Vector3();	
	vList[1]->set(m_brf.x, m_brf.y + height, m_brf.z + length);
	vList[2] = new Vector3();	
	vList[2]->set(m_brf.x + width, m_brf.y + height, m_brf.z + length);
	vList[3] = new Vector3();	
	vList[3]->set(m_brf.x + width, m_brf.y, m_brf.z + length);	
	m_side[5] = new ObstaclePolygon(4, vList, m_id);
  
	Vector3 center(m_brf.x + width/2, m_brf.y + height/2, m_brf.z + length/2);

	Vector3 tempVec;
	tempVec = m_tlb - center;
	double radius = tempVec.length();
	tempVec = m_brf - center;
	double temp = tempVec.length();
	if (temp > radius) radius = temp;

	m_boundingSphere = new ObstacleSphere(center, radius, m_id);

}

Obstacle *ObstacleBox::clone(void) 
{
	return new ObstacleBox(*this);
}

int ObstacleBox::operator==(const ObstacleBox &a) 
{ 
	//return (m_tlb == a.m_tlb && m_brf == a.m_brf);
	return 1;
}

ISectData ObstacleBox::intersectionWithRay(Vector3 &raydirection, Vector3 &rayorigin) 
{
	// for check quickly, use sphere first
	ISectData data;
	data = m_boundingSphere->doesRayIntersect(raydirection, rayorigin);
	if (data.m_intersectionflag == 0) {
		return data;
	}

	ISectData closestIntersect;
	double distToClosestIntersect =  MAX_NUM;

	closestIntersect.m_intersectionflag = 0;
	Vector3 tempvec(0,0,0);
	// find the closest intersection
	for (int i = 0; i < 6; i++) {
		data = m_side[i]->doesRayIntersect(raydirection, rayorigin);
		if (data.m_intersectionflag == 1) { 
			tempvec = data.m_point - rayorigin;
			double temp = tempvec.length();
			if (temp < distToClosestIntersect) {
				int test = m_id;
				distToClosestIntersect = temp;
				closestIntersect = data;
			}
		}
	}

	return closestIntersect;
}

/* ########################################################################
 SPHERE
######################################################################### */
ObstacleSphere::ObstacleSphere(const Vector3 &sphereOrigin, const double sphereRadius, Ogre::SceneManager* mSceneMgr, char* name, char* filename, int Id) 
: Obstacle(mSceneMgr, name, filename, NULL, 0.0f)
{
	init(sphereOrigin, sphereRadius, Id);
}


ObstacleSphere::ObstacleSphere(const Vector3 &sphereOrigin, const double sphereRadius, int Id) 
: Obstacle()
{
	init(sphereOrigin, sphereRadius, Id);
}

ObstacleSphere::~ObstacleSphere(void)
{
}
	
void ObstacleSphere::init(const Vector3 &sphereOrigin, const double sphereRadius, int Id)
{
	if(Id == -1) {
		m_id = m_obstracleCount;
		m_obstacleList.push_back(this);
		m_obstracleCount++;	
		char bName[10];
		sprintf(bName, "obstacle%d", m_id);
		m_name = bName;
		cout << m_name << " is created " << endl;	
	}
	
	m_radius = sphereRadius;
	m_origin = sphereOrigin;
}

Obstacle *ObstacleSphere::clone(void) {
  return new ObstacleSphere(*this);
}   

void ObstacleSphere::update(const Ogre::Vector3 &sphereOrigin) {
	m_origin = sphereOrigin;
}


int ObstacleSphere::operator==(ObstacleSphere &a) { 
	if (m_origin != a.m_origin) return 0;
	if (m_radius != a.m_radius) return 0;
	
	return 1;
}

ISectData ObstacleSphere::intersectionWithRay(Vector3 &raydirection, Vector3 &rayorigin) {
	ISectData data;
	Vector3 rdirection = direction(raydirection);      
	Vector3 tempVec;
	
	if (rayorigin ==m_origin) {
		data.m_intersectionflag = 1; 
		data.m_point = rdirection * m_radius; 
		tempVec = m_origin - data.m_point;
		data.m_normal = direction(tempVec);
		return data; 
	}

	data.m_intersectionflag = 0;  

	Vector3 zero(0,0,0);
	if (rdirection == zero) return data;

	tempVec = m_origin - rayorigin;	
	double v = tempVec.dotProduct(rdirection);

	Vector3 temp = m_origin - rayorigin;
	double d = pow(m_radius, 2) - pow(temp.length(), 2) + pow(v, 2);

	if (d < 0) return data;

	temp = rayorigin - m_origin;
	if (temp.length() < m_radius) {
		temp = rdirection + rayorigin;
		data.m_point = temp * (v + sqrt(d)); // inside 
	}
	else { 	
		temp = rdirection + rayorigin;	
		data.m_point = temp * (v - sqrt(d)); // outside 
	}

	data.m_intersectionflag = 1; 

	data.m_normal = direction(data.m_point - m_origin);	

	return data;

}

/* ########################################################################
 POLYGON
######################################################################### */
ObstaclePolygon::ObstaclePolygon(int numverts, Vector3** vList, SceneManager* mSceneMgr, char* name, char* filename, int Id) 
: Obstacle(mSceneMgr, name, filename, NULL, 0.0f)
{
	init(numverts, vList, Id);
}

ObstaclePolygon::ObstaclePolygon(int numverts, Vector3** vList, int Id) 
: Obstacle()
{
	init(numverts, vList, Id);
}

ObstaclePolygon::ObstaclePolygon(const ObstaclePolygon &p, SceneManager* mSceneMgr, char* name, char* filename) 
: Obstacle(mSceneMgr, name, filename, NULL, 0.0f)
{
	m_d = p.m_d;
	m_i1 = p.m_i1;
	m_i2 = p.m_i2;
	m_normal = p.m_normal;

	m_numvertices = p.m_numvertices;
	m_vertex = new Vector3*[m_numvertices];
	for (int i = 0; i < m_numvertices; i++) 
	{
		m_vertex[i] = new Vector3();
		m_vertex[i] = p.m_vertex[i];
	}
}

ObstaclePolygon::~ObstaclePolygon(void) 
{
	delete[] m_vertex;
}

Ogre::Vector3 ** ObstaclePolygon::getVertex(void)
{
	return m_vertex;
}

void ObstaclePolygon::init(int numverts, Vector3** vList, int Id)
{
	if(Id == -1) {
		m_id = m_obstracleCount;
		m_obstacleList.push_back(this);
		m_obstracleCount++;	
		char bName[10];
		sprintf(bName, "obstacle%d", m_id);
		m_name = bName;
		//cout << m_name << " is created " << endl;	
	} 
	
	m_numvertices = numverts;
	m_vertex = vList;
	
	m_normal = *m_vertex[1];
	m_normal -= *m_vertex[0];
	Vector3 tempVec = *m_vertex[2];
	tempVec -= *m_vertex[0];
	m_normal = m_normal.crossProduct(tempVec); 
	m_normal.normalise();

	tempVec.set(-m_vertex[0]->x,-m_vertex[0]->y,-m_vertex[0]->z); 
	m_d = tempVec.dotProduct(m_normal); 

	// find dominant axis of normal vector
	if (fabs(m_normal.x) > fabs(m_normal.y) && fabs(m_normal.x) > fabs(m_normal.z)) {
		m_i1 = 1; 
		m_i2 = 2;
	}
	else if (fabs(m_normal.y) >= fabs(m_normal.x) && fabs(m_normal.y) >= fabs(m_normal.z)) {
		m_i1 = 0; 
		m_i2 = 2;    
	}
	else {
		m_i1 = 0; 
		m_i2 = 1;   
	}

}

Obstacle *ObstaclePolygon::clone(void) 
{
	return new ObstaclePolygon(*this);
}

void ObstaclePolygon::update(void)
{
	// normal to poly
	m_normal = *m_vertex[1];
	m_normal -= *m_vertex[0];
	Vector3 tempVec = *m_vertex[2];
	tempVec -= *m_vertex[0];
	m_normal = m_normal.crossProduct(tempVec); 
	m_normal.normalise();

	tempVec.set(-m_vertex[0]->x,-m_vertex[0]->y,-m_vertex[0]->z); 
	m_d = tempVec.dotProduct(m_normal); 

	// find dominant axis of normal vector
	if (fabs(m_normal.x) > fabs(m_normal.y) && fabs(m_normal.x) > fabs(m_normal.z)) {
		m_i1 = 1; 
		m_i2 = 2;
	}
	else if (fabs(m_normal.y) >= fabs(m_normal.x) && fabs(m_normal.y) >= fabs(m_normal.z)) {
		m_i1 = 0; 
		m_i2 = 2;    
	}
	else {
		m_i1 = 0; 
		m_i2 = 1;   
	}
}


int ObstaclePolygon::operator==(const ObstaclePolygon &a) 
{ 
	int i, test;

	if (m_numvertices != a.m_numvertices) return 0; // definitely different

	for (i=0, test=1; test && (i < m_numvertices); i++)       
		test = test && (m_vertex[i]==a.m_vertex[i]);

	return(test);
}

ISectData ObstaclePolygon::intersectionWithRay(Vector3 &raydirection, Vector3 &rayorigin) 
{
	ISectData data;
	data.m_intersectionflag = 0;  

	Vector3 rdirection = direction(raydirection);    

	if (rdirection.x == 0 && rdirection.y == 0 && rdirection.z == 0)
		return data;
  
	double denom = m_normal.dotProduct(rdirection);

	// if polygon is parallel to the ray, or contains the ray,
	if (denom == 0) return(data); 

	double t = - (m_d + m_normal.dotProduct(rayorigin)) / denom;

	// if intersection is behind the ray's origin, 
	if (t < 0) return (data);

	// calculate the point of intersection 
	data.m_point = rayorigin + rdirection*t;

	// determine if the point of intersection is inside the polygon 
	double u0 = data.m_point[m_i1] - m_vertex[0]->get(m_i1);
	double v0 = data.m_point[m_i2] - m_vertex[0]->get(m_i2);	
	double u1, u2, v1, v2, alpha, beta;
	char inside = 0;  
	int  i = 2;      

    // break the polygon up into triangles and check each one to see if the
    // point of intersection is within it.
    do {
		// The polygon is viewed as (n-2) triangles. 
		u1 = m_vertex[i-1]->get(m_i1) - m_vertex[0]->get(m_i1); 
		v1 = m_vertex[i-1]->get(m_i2) - m_vertex[0]->get(m_i2);
		u2 = m_vertex[i]->get(m_i1) - m_vertex[0]->get(m_i1); 
		v2 = m_vertex[i]->get(m_i2) - m_vertex[0]->get(m_i2);		
    
		if (u1 == 0) {
			beta = u0/u2;
			if ((beta >= 0.0) && (beta <= 1.0)) {
				alpha = (v0 - beta*v2)/v1;
				inside = ((alpha >= 0.0) && (alpha+beta <= 1.0));
			}
		} 
		else {
			beta = (v0*u1 - u0*v1)/(v2*u1 - u2*v1);
			if ((beta >= 0.0)&&(beta <= 1.0)) { 
				alpha = (u0 - beta*u2)/u1;
				inside = ((alpha >= 0.0)&&((alpha+beta) <= 1.0));
			}
		}
		
		i++;
	} while ((inside == 0) && (i < m_numvertices));
  

	data.m_intersectionflag = inside; 

	data.m_normal = m_normal;

	return(data);  
}


