// SimObject.cpp : Base class for all object components 
// Animation Class CS527
// Hyejung Hur

#include "SimObject.h"

using namespace std;
using namespace Ogre;


SimObject::SimObject(void) :
m_node(NULL), m_name(""), m_id(-1), m_lastUpdate(0), m_mass(9.07),
m_dimEntity(NULL), m_geomEntity(NULL), m_drawMode(NORMAL_MODE), m_active(false),
m_dimNode(NULL), m_geomNode(NULL), m_aniGeoNode(NULL), m_aniGeomEntity(NULL)
{
	m_roll = m_pitch = m_yaw = 0;
}

SimObject::SimObject(SceneManager* mSceneMgr, char* name, char* filename, char* anifilename, double gab) :
m_node(NULL), m_name(""), m_id(-1), m_lastUpdate(0), m_mass(9.07), 
m_drawMode(NORMAL_MODE), m_active(false),
m_dimEntity(NULL), m_geomEntity(NULL), m_dimNode(NULL), m_geomNode(NULL), 
m_aniGeoNode(NULL), m_aniGeomEntity(NULL)
{
	if(name!= NULL)
		m_name = name;
	m_roll = m_pitch = m_yaw = 0;
	if(mSceneMgr != NULL && name != NULL && filename != NULL)
	{
		m_node = mSceneMgr->getRootSceneNode()->createChildSceneNode();
		m_geomNode = m_node->createChildSceneNode();
		m_geomEntity = mSceneMgr->createEntity( name, filename );
		m_geomNode->attachObject( m_geomEntity );

		char dimname[15];

		if(anifilename)
		{
			sprintf(dimname, "%s-ani", name);
			m_aniGeoNode = m_geomNode->createChildSceneNode();
			m_aniGeomEntity = mSceneMgr->createEntity( dimname, anifilename );
			m_aniGeoNode->attachObject( m_aniGeomEntity );
			m_aniGeoNode->setPosition(0.0f, gab, 0.0f);
		}

		sprintf(dimname, "%s-dim", name);
		m_dimNode = m_node->createChildSceneNode();
		m_dimEntity = mSceneMgr->createEntity( dimname, Ogre::SceneManager::PT_CUBE);
		//m_node->attachObject( m_dimEntity );
		// m_node->detachObject(pEnt);
		// setVisible 
	}
}

void SimObject::initScene(bool update)
{
	if(update == true && (m_geomEntity != NULL))
	{
		AxisAlignedBox bound = m_geomEntity->getBoundingBox();
		Vector3 size = bound.getSize();
		if(m_aniGeomEntity)
		{
			AxisAlignedBox anibound = m_aniGeomEntity->getBoundingBox(); 
			Vector3 anisize = anibound.getSize();
			size.y += anisize.y;
			if(anisize.x > size.x)
			{
				size.x = anisize.x;
			}
			if(anisize.z > size.z)
			{
				size.z = anisize.z;
			}
		}
		//size /= 100;
		Vector3 scalefactor(m_dimensions.x/size.x, m_dimensions.y/size.y, m_dimensions.z/size.z);
		m_geomNode->setScale(scalefactor);
	}

	if(m_dimEntity != NULL)
	{
		//Ogre::MeshPtr dimGeom = m_dimEntity->getMesh();
		//Ogre::Box box = dimGeom.get();
		AxisAlignedBox bound = m_dimEntity->getBoundingBox();
		Vector3 size = bound.getSize();
		//size /= 100;
		Vector3 scalefactor(m_dimensions.x/size.x, m_dimensions.y/size.y, m_dimensions.z/size.z);
		m_dimNode->setScale(scalefactor);

		//std::cout << "geom size " << size.x << " " << size.y << " " << size.z << std::endl;
		//std::cout << "dimention " << m_dimensions.x << " " << m_dimensions.y << " " <<  m_dimensions.z << std::endl;
	}
}

SimObject::~SimObject(void) 
{
}

void SimObject::setDimensionDrawMode(int mode)
{
	if(m_drawMode != mode)
	{
		switch(m_drawMode)
		{
		case NORMAL_MODE :
			if(m_geomEntity)
				m_geomNode->detachObject(m_geomEntity);
			break;
		case DIMENSION_MODE :
			if(m_dimEntity)
				m_dimNode->detachObject(m_dimEntity);
			break;
		case ALL_SHOW_MODE :
			if(m_geomEntity)
			m_geomNode->detachObject(m_geomEntity);
			if(m_dimEntity)
				m_dimNode->detachObject(m_dimEntity);
			break;
		}

		m_drawMode = mode;
		switch(m_drawMode)
		{
		case NORMAL_MODE :
			if(m_geomEntity)
				m_geomNode->attachObject(m_geomEntity);
			break;
		case DIMENSION_MODE :
			if(m_dimEntity)
				m_dimNode->attachObject(m_dimEntity);
			break;
		case ALL_SHOW_MODE :
			if(m_geomEntity)
				m_geomNode->attachObject(m_geomEntity);
			if(m_dimEntity)
				m_dimNode->attachObject(m_dimEntity);
			break;
		case ALL_HIDE_MODE :
			if(m_geomEntity)
				m_geomNode->detachObject(m_geomEntity);
			if(m_dimEntity)
				m_dimNode->detachObject(m_dimEntity);
			break;
		}
	}
	
}